"""
RelevanceAI Operations wrappers for use from a Dataset object

.. code-block::

    from relevanceai import Client

    client = Client()

    dataset = client.Dataset()

    dataset.vectorize_text(*args, **kwargs)

    dataset.reduce_dims(*args, **kwargs)

    dataset.cluster(*args **kwarsgs)
"""
import pandas as pd
from tqdm.auto import tqdm
from typing import Any, Dict, List, Optional
from datetime import datetime
from relevanceai.dataset.write import Write

from relevanceai.utils.decorators.analytics import track
from relevanceai.constants import EXPLORER_APP_LINK


def get_ptp_args() -> List[str]:
    """
    Returns all arguments in the PullTransformPush.__init__ func
    """

    from relevanceai.operations_new.ops_run import PullTransformPush, arguments

    return arguments(PullTransformPush)


class Operations(Write):
    @staticmethod
    def _get_filters(input_fields: List[str], output_fields: List[str]):
        """
        Creates the filters necessary to search all documents
        within a dataset that contain fields specified in "input_fields"
        but do not contain the fields specified in "output_fields"

        e.g.
        fields = ["text", "title"]
        vector_fields = ["text_use_vector_", "title_use_vector_"]

        we want to search the dataset where:
        ("text" * ! "text_use_vector_") + ("title" * ! "title_use_vector_")

        Since the current implementation of filtering only accounts for CNF and not DNF boolean logic,
        We must use boolean algebra here to obtain the CNF from a DNF expression.

        CNF = Conjunctive Normal Form (Sum of Products)
        DNF = Disjunctive Normal Form (Product of Sums)

        This means converting the above to:
        ("text" + "title") * ("text" + ! "title_use_vector_") *
        (! "text_use_vector_" + "title") * (! "text_use_vector_" + ! "title_use_vector_")

        Arguments:
            input_fields: List[str]
                A list of fields within the dataset

            output_fields: List[str]
                A list of output_fields, created from the operation.
                These would be present in processed documents

        Returns:
            filters: List[Dict[str, Any]]
                A list of filters.
        """

        if len(input_fields) > 1:
            iters = len(input_fields) ** 2

            filters: list = []
            for i in range(iters):
                binary_array = [character for character in str(bin(i))][2:]
                mixed_mask = ["0"] * (
                    len(input_fields) - len(binary_array)
                ) + binary_array
                mask = [int(value) for value in mixed_mask]
                # Creates a binary mask the length of fields provided
                # for two fields, we need 4 iters, going over [(0, 0), (1, 0), (0, 1), (1, 1)]

                condition_value = [
                    {
                        "field": field if mask[index] else vector_field,
                        "filter_type": "exists",
                        "condition": "==" if mask[index] else "!=",
                        "condition_value": "",
                    }
                    for index, (field, vector_field) in enumerate(
                        zip(input_fields, output_fields)
                    )
                ]
                filters += [{"filter_type": "or", "condition_value": condition_value}]

        else:  # Special Case when only 1 field is provided
            condition_value = [
                {
                    "field": input_fields[0],
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": " ",
                },
                {
                    "field": output_fields[0],
                    "filter_type": "exists",
                    "condition": "!=",
                    "condition_value": " ",
                },
            ]
            filters = condition_value

        return filters

    @track
    def reduce_dims(
        self,
        vector_fields: List[str],
        n_components: int = 3,
        batched: bool = False,
        model: Optional[Any] = None,
        model_kwargs: Optional[dict] = None,
        alias: Optional[str] = None,
        filters: Optional[list] = None,
        chunksize: Optional[int] = 100,
        output_field: str = None,
        **kwargs,
    ):
        """It takes a list of fields, a list of models, a list of filters, and a chunksize, and then runs
        the DimReductionOps class on the documents in the dataset

        Parameters
        ----------
        fields : List[str]
            List[str]
        models : Optional[List[Any]]
            List[Any] = None,
        filters : Optional[List[Dict[str, Any]]]
            A list of dictionaries, each dictionary containing a filter.
        chunksize : int, optional
            The number of documents to process at a time.

        Returns
        -------
            Nothing is being returned.

        """
        from relevanceai.operations_new.dr.ops import DimReductionOps

        model = "pca" if model is None else model

        ops = DimReductionOps(
            credentials=self.credentials,
            vector_fields=vector_fields,
            n_components=n_components,
            model=model,
            model_kwargs=model_kwargs,
            alias=alias,
            output_field=output_field,
        )

        filters = [] if filters is None else filters
        filters += Operations._get_filters(
            input_fields=vector_fields,
            output_fields=[ops.model.vector_name(vector_fields, output_field)],
        )

        res = ops.run(
            dataset=self,
            select_fields=vector_fields,
            chunksize=chunksize,
            filters=filters,
            batched=batched,
            **kwargs,
        )

        return ops

    @track
    def vectorize_text(
        self,
        fields: List[str],
        batched: bool = True,
        models: Optional[List[Any]] = None,
        filters: Optional[list] = None,
        chunksize: Optional[int] = None,
        output_fields: list = None,
        **kwargs,
    ):
        """It takes a list of fields, a list of models, a list of filters, and a chunksize, and then it runs
        the VectorizeOps function on the documents in the database

        Parameters
        ----------
        fields : List[str]
            List[str]
        models : List[Any]
            List[Any]
        filters : List[Dict[str, Any]]
            List[Dict[str, Any]]
        chunksize : int, optional
            int = 100,

        Returns
        -------
            Nothing

        """
        from relevanceai.operations_new.vectorize.text.ops import VectorizeTextOps

        models = ["all-mpnet-base-v2"] if models is None else models

        ops = VectorizeTextOps(
            credentials=self.credentials,
            fields=fields,
            models=models,
            output_fields=output_fields,
        )

        filters = [] if filters is None else filters
        filters += Operations._get_filters(
            input_fields=fields, output_fields=ops.vector_fields
        )

        res = ops.run(
            dataset=self,
            select_fields=fields,
            filters=filters,
            batched=batched,
            chunksize=chunksize,
            output_fields=output_fields,
            **kwargs,
        )

        return ops

    @track
    def vectorize_image(
        self,
        fields: List[str],
        models: Optional[List[Any]] = None,
        batched: Optional[bool] = True,
        filters: Optional[list] = None,
        chunksize: Optional[int] = 20,
        **kwargs,
    ):
        """It takes a list of fields, a list of models, a list of filters, and a chunksize, and then it runs
        the VectorizeOps function on the documents in the database

        Parameters
        ----------
        fields : List[str]
            List[str]
        models : List[Any]
            List[Any]
        filters : List[Dict[str, Any]]
            List[Dict[str, Any]]
        chunksize : int, optional
            int = 100,

        Returns
        -------
            Nothing

        """
        from relevanceai.operations_new.vectorize.image.ops import VectorizeImageOps

        models = ["clip"] if models is None else models

        ops = VectorizeImageOps(
            credentials=self.credentials,
            fields=fields,
            models=models,
        )

        filters = [] if filters is None else filters
        filters += Operations._get_filters(
            input_fields=fields, output_fields=ops.vector_fields
        )

        res = ops.run(
            dataset=self,
            select_fields=fields,
            filters=filters,
            batched=batched,
            chunksize=chunksize,
            **kwargs,
        )

        return ops

    @track
    def label(
        self,
        vector_fields: List[str],
        label_documents: List[Any],
        expanded: bool = True,
        max_number_of_labels: int = 1,
        similarity_metric: str = "cosine",
        similarity_threshold: float = 0,
        label_field: str = "label",
        label_vector_field: str = "label_vector_",
        batched: bool = True,
        filters: Optional[list] = None,
        chunksize: Optional[int] = 100,
        output_field: str = None,
        **kwargs,
    ):
        """This function takes a list of documents, a list of vector fields, and a list of label documents,
        and then it labels the documents with the label documents

        Parameters
        ----------
        vector_fields : List[str]
            List[str]
        label_documents : List[Any]
            List[Any]
        expanded, optional
            If True, the label_vector_field will be a list of vectors. If False, the label_vector_field
        will be a single vector.
        max_number_of_labels : int, optional
            int = 1,
        similarity_metric : str, optional
            str = "cosine",
        filters : Optional[list]
            A list of filters to apply to the documents.
        chunksize : int, optional
            The number of documents to process at a time.
        similarity_threshold : float, optional
            float = 0,
        label_field : str, optional
            The name of the field that will contain the label.
        label_vector_field, optional
            The field that will be added to the documents that contain the label vector.

        Returns
        -------
            The return value is a list of documents.

        """

        from relevanceai.operations_new.label.ops import LabelOps

        if len(vector_fields) > 1:
            raise ValueError(
                "We currently do not support on more than 1 vector length."
            )

        # self.datasets.create(
        #     dataset_id=self.dataset_id, schema={output_field: "chunks"}
        # )

        ops = LabelOps(
            credentials=self.credentials,
            label_documents=label_documents,
            vector_field=vector_fields[0],
            expanded=expanded,
            max_number_of_labels=max_number_of_labels,
            similarity_metric=similarity_metric,
            similarity_threshold=similarity_threshold,
            label_field=label_field,
            label_vector_field=label_vector_field,
            output_field=output_field,
        )
        # Add an exists filter
        if filters is None:
            filters = []

        filters += [
            {
                "field": vector_fields[0],
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
        ]
        # Check if output field already exists
        res = ops.run(
            dataset=self,
            filters=filters,
            batched=batched,
            chunksize=chunksize,
            select_fields=vector_fields,
            **kwargs,
        )

        return ops

    @track
    def label_from_dataset(
        self,
        vector_fields: list,
        label_dataset,
        max_number_of_labels: int = 1,
        label_vector_field="label_vector_",
        expanded: bool = False,
        similarity_metric: str = "cosine",
        label_field: str = "label",
        batched: bool = True,
        filters: list = None,
        similarity_threshold=0.1,
        chunksize: int = 100,
        output_field: str = None,
        **kwargs,
    ):
        """
        Label from another dataset
        """
        if output_field is None:
            output_field = "_label_." + label_dataset.dataset_id + "." + label_field
        label_documents = label_dataset.get_all_documents()
        return self.label(
            vector_fields=vector_fields,
            label_documents=label_documents,
            expanded=expanded,
            output_field=output_field,
            max_number_of_labels=max_number_of_labels,
            similarity_metric=similarity_metric,
            similarity_threshold=similarity_threshold,
            label_field=label_field,
            label_vector_field=label_vector_field,
            batched=batched,
            filters=filters,
            chunksize=chunksize,
            **kwargs,
        )

    @track
    def split_sentences(
        self,
        text_fields: List[str],
        output_field="_splittextchunk_",
        language: str = "en",
        inplace: bool = True,
        batched: bool = False,
        filters: Optional[list] = None,
        chunksize: Optional[int] = 100,
        **kwargs,
    ):
        """
        This function splits the text in the `text_field` into sentences and stores the sentences in
        the `output_field`

        Parameters
        ----------
        text_field : str
            The field in the documents that contains the text to be split into sentences.
        output_field, optional
            The name of the field that will contain the split sentences.
        language : str, optional
            The language of the text. This is used to determine the sentence splitting rules.

        """
        from relevanceai.operations_new.processing.text.sentence_splitting.ops import (
            SentenceSplitterOps,
        )

        ops = SentenceSplitterOps(
            credentials=self.credentials,
            text_fields=text_fields,
            language=language,
            inplace=inplace,
            output_field=output_field,
        )

        filters = [] if filters is None else filters

        res = ops.run(
            dataset=self,
            select_fields=text_fields,
            batched=batched,
            filters=filters,
            chunksize=chunksize,
            **kwargs,
        )

        return ops

    @track
    def cluster(
        self,
        vector_fields: List[str],
        model: Optional[Any] = None,
        alias: Optional[str] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
        chunksize: Optional[int] = 128,
        filters: Optional[list] = None,
        batched: Optional[bool] = False,
        include_cluster_report: bool = False,
        **kwargs,
    ):
        """`cluster` is a function that takes in a list of vector fields, a model, an alias, a list of
        filters, a boolean value, a dictionary of model keyword arguments, and a list of keyword
        arguments. It returns an object of type `ClusterOps`

        Example
        ----------
        .. code-block::

            from sklearn.cluster import KMeans
            model = KMeans()

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            cluster_ops = ds.cluster(
                model=model, vector_fields=["sample_vector_"],
                alias="kmeans-8"
            )

        Parameters
        ----------
        vector_fields : List[str]
            A list of possible vector fields
        model : Optional[Any]
            The clustering model to use. Currently, we support KMeans and MiniBatchKMeans.
        alias : Optional[str]
            The name of the cluster model.
        filters : Optional[list]
            Optional[list] = None,
        include_cluster_report : bool, optional
            bool = True
        model_kwargs : Optional[Dict[str, Any]]
            The cluster config to use
            You can change the number of clusters for kmeans using:
            `cluster_config={"n_clusters": 10}`. For a full list of
            possible parameters for different models, simply check how
            the cluster models are instantiated.

        Returns
        -------
            The cluster object

        """

        from relevanceai.operations_new.cluster.ops import ClusterOps

        model = "kmeans" if model is None else model
        model_kwargs = {} if model_kwargs is None else model_kwargs

        ops = ClusterOps(
            model=model,
            alias=alias,  # type: ignore
            vector_fields=vector_fields,  # type: ignore
            verbose=False,
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            model_kwargs=model_kwargs,
            include_cluster_report=include_cluster_report,
            **kwargs,
        )

        if filters is not None:
            filters = ops._get_filters(filters, vector_fields)

        ops.run(
            dataset=self,
            select_fields=vector_fields,
            batched=batched,
            chunksize=chunksize,
            filters=filters,
        )
        print()
        print(
            f"""You can now utilise the ClusterOps object using the below:

    cluster_ops = client.ClusterOps(
        alias='{ops.alias}',
        vector_fields={ops.vector_fields},
        dataset_id='{self.dataset_id}'
    )"""
        )
        print()
        print("Configure your new explore app below:")
        print(EXPLORER_APP_LINK.format(self.dataset_id))
        return ops

    def _get_alias(self, alias: Any) -> str:
        # Auto-generates alias here
        if alias is None:
            if hasattr(self.model, "n_clusters"):
                n_clusters = (
                    self.n_clusters
                    if self.n_clusters is not None
                    else self.model.n_clusters
                )
                alias = f"{self.model_name}-{n_clusters}"

            elif hasattr(self.model, "k"):
                n_clusters = (
                    self.n_clusters if self.n_clusters is not None else self.model.k
                )
                alias = f"{self.model_name}-{n_clusters}"

            else:
                alias = self.model_name

            Warning.MISSING_ALIAS.format(alias=alias)  # type: ignore

        if self.verbose:
            print(f"The alias is `{alias.lower()}`.")
        return alias.lower()

    @track
    def batch_cluster(
        self,
        vector_fields: List[str],
        model: Any = None,
        alias: Optional[str] = None,
        filters: Optional[list] = None,
        model_kwargs: Dict = None,
        chunksize: int = 128,
        **kwargs,
    ):
        from relevanceai.operations_new.cluster.batch.ops import BatchClusterOps

        n_clusters = ({} if model_kwargs is None else model_kwargs).get("n_clusters", 8)
        if kwargs.get("transform_chunksize") is None:
            kwargs["transform_chunksize"] = max(n_clusters * 5, 128)

        run_kwargs = {}
        for key in get_ptp_args():
            if key in kwargs:
                run_kwargs[key] = kwargs.pop(key)

        if "credentials" not in kwargs:
            kwargs["credentials"] = self.credentials

        cluster_ops = BatchClusterOps(
            model=model,
            alias=alias,
            dataset_id=self.dataset_id,
            vector_fields=vector_fields,
            model_kwargs=model_kwargs,
            **kwargs,
        )

        filters = cluster_ops._get_filters(filters, vector_fields)  # type: ignore

        cluster_ops.run(self, filters=filters, chunksize=chunksize, **run_kwargs)

        return cluster_ops

    def extract_sentiment(
        self,
        text_fields: List[str],
        model_name: str = "cardiffnlp/twitter-roberta-base-sentiment",
        highlight: bool = False,
        max_number_of_shap_documents: int = 1,
        min_abs_score: float = 0.1,
        sensitivity: float = 0,
        filters: Optional[list] = None,
        output_fields: list = None,
        chunksize: int = 128,
        batched: bool = True,
        **kwargs,
    ):
        """
        Extract sentiment from the dataset

        If you are dealing with news sources, you will want
        more sensitivity, as more news sources are likely to be neutral

        """
        from relevanceai.operations_new.sentiment.ops import SentimentOps

        ops = SentimentOps(
            credentials=self.credentials,
            text_fields=text_fields,
            model_name=model_name,
            highlight=highlight,
            max_number_of_shap_documents=max_number_of_shap_documents,
            min_abs_score=min_abs_score,
            output_fields=output_fields,
            sensitivity=sensitivity,
        )
        filters = [] if filters is None else filters
        filters += SentimentOps._get_filters(text_fields, model_name)
        ops.run(
            self,
            filters=filters,
            select_fields=text_fields,
            chunksize=chunksize,
            batched=batched,
            **kwargs,
        )
        return ops

    def extract_emotion(
        self,
        text_fields: list,
        model_name="joeddav/distilbert-base-uncased-go-emotions-student",
        filters: list = None,
        chunksize: int = 100,
        output_fields: list = None,
        min_score: float = 0.3,
        batched: bool = True,
        refresh: bool = False,
        **kwargs,
    ):
        """
        Extract an emotion.

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            ds.extract_emotion(
                text_fields=["sample_1_label"],
            )

        """
        from relevanceai.operations_new.emotion.ops import EmotionOps

        filters = [] if filters is None else filters
        ops = EmotionOps(
            credentials=self.credentials,
            text_fields=text_fields,
            model_name=model_name,
            output_fields=output_fields,
            min_score=min_score,
        )
        filters += [
            {
                "field": text_field,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for text_field in text_fields
        ]
        ops.run(
            self,
            filters=filters,
            select_fields=text_fields,
            chunksize=chunksize,
            batched=batched,
            output_fields=output_fields,
            refresh=refresh,
            **kwargs,
        )

        return ops

    def apply_transformers_pipeline(
        self,
        text_fields: list,
        pipeline,
        output_fields: Optional[List[str]] = None,
        filters: Optional[list] = None,
        refresh: bool = False,
        **kwargs,
    ):
        """
        Apply a transformers pipeline generically.

        .. code-block::

            from transformers import pipeline
            pipeline = pipeline("automatic-speech-recognition", model="facebook/wav2vec2-base-960h", device=0)
            ds.apply_transformers_pipeline(
                text_fields, pipeline
            )

        """
        from relevanceai.operations_new.processing.transformers.ops import (
            TransformersPipelineOps,
        )

        ops = TransformersPipelineOps(
            text_fields=text_fields,
            pipeline=pipeline,
            output_fields=output_fields,
            credentials=self.credentials,
        )
        ops.run(
            self,
            filters=filters,
            select_fields=text_fields,
            output_fields=output_fields,
            refresh=refresh,
            **kwargs,
        )
        return ops

    def scale(
        self,
        vector_fields: List[str],
        model: Optional[str] = "standard",
        alias: Optional[str] = None,
        model_kwargs: Optional[dict] = None,
        filters: Optional[list] = None,
        batched: Optional[bool] = None,
        chunksize: Optional[int] = None,
        **kwargs,
    ):

        from relevanceai.operations_new.scaling.ops import ScaleOps

        chunksize = chunksize if batched is None else batched
        filters = [] if filters is None else filters
        batched = False if batched is None else batched

        ops = ScaleOps(
            vector_fields=vector_fields,
            model=model,
            alias=alias,
            model_kwargs=model_kwargs,
        )

        ops.run(
            dataset=self,
            batched=batched,
            chunksize=chunksize,
            filters=filters,
            select_fields=vector_fields,
            **kwargs,
        )
        return ops

    def subcluster(
        self,
        vector_fields: List[str],
        alias: str,
        parent_field: str,
        model: Any = "kmeans",
        cluster_field: str = "_cluster_",
        model_kwargs: Optional[dict] = None,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        min_parent_cluster_size: int = 0,
        **kwargs,
    ):
        from relevanceai.operations_new.cluster.sub.ops import SubClusterOps

        ops = SubClusterOps(
            model=model,
            alias=alias,
            vector_fields=vector_fields,
            parent_field=parent_field,
            model_kwargs=model_kwargs,
            cluster_field=cluster_field,
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            cluster_ids=cluster_ids,
            min_parent_cluster_size=min_parent_cluster_size,
            **kwargs,
        )

        run_kwargs = {}
        for key in get_ptp_args():
            if key in kwargs:
                run_kwargs[key] = kwargs.pop(key)

        # Building an infinitely hackable SDK

        # Add filters and select fields
        select_fields = vector_fields + [parent_field]
        if filters is None:
            filters = []

        if cluster_ids is not None:
            filters += [
                {
                    "field": parent_field,
                    "filter_type": "exact_match",
                    "condition": "==",
                    "condition_value": cluster_id,
                }
                for cluster_id in cluster_ids
            ]
        filters += [
            {
                "field": vf,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vf in vector_fields
        ]
        filters += [
            {
                "field": parent_field,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
        ]

        ops.run(
            self,
            filters=filters,
            select_fields=select_fields,
            **run_kwargs,
        )
        print(
            f"""You can now utilise the ClusterOps object based on subclustering.

    cluster_ops = client.ClusterOps(
        alias='{ops.alias}',
        vector_fields={ops.vector_fields},
        dataset_id='{self.dataset_id}'
    )"""
        )

        from relevanceai.operations_new.cluster.ops import ClusterOps

        model = "kmeans" if model is None else model
        model_kwargs = {} if model_kwargs is None else model_kwargs

        ops = ClusterOps(
            model=model,
            alias=alias,  # type: ignore
            vector_fields=vector_fields,  # type: ignore
            verbose=False,
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            model_kwargs=model_kwargs,
            **kwargs,
        )
        return ops

    def byo_cluster(
        self,
        vector_fields: list,
        alias: str,
        byo_cluster_field: str,
        centroids: list = None,
    ):
        """
        Bring your own clusters and we can calculate the centroids for you.

        Example
        =========

        .. code-block::

            dataset = client.Dataset("retail_reviews")
            cluster_ops = dataset.byo_cluster(
                vector_fields=['reviews.title_mpnet_vector_'],
                alias="manufacturer_two",
                byo_cluster_field="manufacturer"
            )

        """
        from relevanceai.operations_new.cluster.ops import ClusterOps

        ops = ClusterOps(
            model=None,
            alias=alias,
            verbose=False,
            vector_fields=vector_fields,
            credentials=self.credentials,
            dataset_id=self.dataset_id,
            byo_cluster_field=byo_cluster_field,
        )
        # here we create the centroids for the clusters
        if centroids is None:
            results = ops.create_centroids()
        else:
            results = self.datasets.cluster.centroids.insert(
                dataset_id=self.dataset_id,
                cluster_centers=centroids,
                vector_fields=vector_fields,
                alias=alias,
            )
        return ops

    def clean_text(
        self,
        text_fields: list,
        output_fields: list = None,
        remove_html_tags: bool = True,
        lower=False,
        remove_punctuation=True,
        remove_digits=True,
        remove_stopwords: list = None,
        lemmatize: bool = False,
        filters: list = None,
        replace_words: dict = None,
        **kwargs,
    ):
        """
        Cleans text for you!
        """
        from relevanceai.operations_new.processing.text.clean.ops import CleanTextOps

        if output_fields is None:
            output_fields = [t + "_clean" for t in text_fields]
            print(f"The output fields are {output_fields}.")

        ops = CleanTextOps(
            credentials=self.credentials,
            text_fields=text_fields,
            output_fields=output_fields,
            remove_html_tags=remove_html_tags,
            lower=lower,
            remove_punctuation=remove_punctuation,
            remove_digits=remove_digits,
            remove_stopword=remove_stopwords,
            lemmatize=lemmatize,
            replace_words=replace_words,
            **kwargs,
        )

        print("🥸 A clean house is a sign of no Internet connection.")
        ops.run(
            self,
            filters=filters,
            select_fields=text_fields,
            batched=True,
            output_fields=output_fields,
        )

        return ops

    def count_text(
        self,
        text_fields: list,
        count_words: bool = True,
        count_characters: bool = True,
        count_sentences: bool = True,
        filters: list = None,
        chunksize: int = 1000,
        refresh: bool = False,
        **kwargs,
    ):
        from relevanceai.operations_new.processing.text.count.ops import CountTextOps

        ops = CountTextOps(
            credentials=self.credentials,
            text_fields=text_fields,
            include_char_count=count_characters,
            include_word_count=count_words,
            include_sentence_count=count_sentences,
        )
        res = ops.run(
            dataset=self,
            select_fields=text_fields,
            chunksize=chunksize,
            filters=filters,
            batched=True,
            refresh=refresh,
            **kwargs,
        )

        return ops

    def analyze_text(
        self,
        fields: list,
        vector_fields: list = None,
        vectorize=True,
        vectorize_models: list = None,
        cluster: bool = True,
        cluster_model=None,
        cluster_alias: str = None,
        subcluster: bool = True,
        subcluster_model=None,
        subcluster_alias: str = None,
        subcluster_parent_field: str = None,
        extract_sentiment: bool = True,
        extract_emotion: bool = False,
        count: bool = True,
        verbose: bool = False,
        filters: list = None,
    ):
        # is it worth separating
        # analyze text and analyze text vectors?
        if verbose:
            print("⚛️ Why can't you trust atoms?")
            print("Because they make up everything!")

        if vectorize:
            if vector_fields is None:
                vector_fields = [text_field + "_vector_" for text_field in fields]
                print(f"Outputting to: {vector_fields}")
            self.vectorize_text(
                fields=fields,
                models=vectorize_models,
                filters=filters,
                output_fields=vector_fields,
            )

        try:
            # Runs clustering and subclustering first
            self.analyze_vectors(
                vector_fields=vector_fields,
                cluster=cluster,
                cluster_model=cluster_model,
                cluster_alias=cluster_alias,
                subcluster=subcluster,
                subcluster_alias=subcluster_alias,
                subcluster_parent_field=subcluster_parent_field,
                subcluster_model=subcluster_model,
                filters=filters,
            )
        except:
            pass

        if extract_emotion:
            try:
                print("Extracting emotion...")
                raise NotImplementedError("Have not implemented emotion yet")
            except:
                pass

        if extract_sentiment:
            print("Extracting sentiment...")
            try:
                self.extract_sentiment(text_fields=fields, filters=filters)
            except:
                pass

        if count:
            try:
                print("Extracting count...")
                self.count_text(
                    text_fields=fields,
                    count_words=True,
                    count_characters=True,
                    count_sentences=True,
                    filters=filters,
                )
            except:
                pass
        # TODO:
        # Launch an explorer app with the right settings
        return

    def analyze_vectors(
        self,
        vector_fields: list = None,  # These vector fields will be used throughout
        cluster: bool = False,
        cluster_model=None,
        cluster_alias: str = None,
        subcluster: bool = False,
        subcluster_alias: str = None,
        subcluster_parent_field: str = None,
        subcluster_model=None,
        filters: list = None,
    ):
        # is it worth separating
        # analyze text and analyze text vectors?
        if cluster:
            if vector_fields is None or cluster_model is None:
                raise ValueError(
                    "Vector fields and cluster_models need to not be None."
                )
            self.cluster(
                vector_fields=vector_fields,
                model=cluster_model,
                filters=filters,
                alias=cluster_alias,
            )

        if subcluster:
            # How do I get the cluster parent field
            # TODO - how do you set the alias and parent field?
            if (
                vector_fields is None
                or subcluster_alias is None
                or subcluster_parent_field is None
                or subcluster_model is None
            ):
                raise ValueError(
                    "Vector fields and subcluster_models and subcluster_parent_field and subcluster_alias need to not be None."
                )
            self.subcluster(
                vector_fields=vector_fields,
                alias=subcluster_alias,
                parent_field=subcluster_parent_field,
                model=subcluster_model,
                filters=filters,
            )

    def extract_keywords(
        self,
        fields: list,
        model_name: str = "all-mpnet-base-v2",
        output_fields: list = None,
        lower_bound: int = 0,
        upper_bound: int = 3,
        chunksize: int = 200,
        max_keywords: int = 1,
        stop_words: list = None,
        filters: list = None,
        batched: bool = True,
        use_maxsum: bool = False,
        nr_candidates: int = 20,
        use_mmr=True,
        diversity=0.7,
        **kwargs,
    ):
        """
        Extract the keyphrases of a text field and output and store it into
        a separate field. This can be used to better explain sentiment,
        label and identify why certain things were clustered together!
        """
        from relevanceai.operations_new.processing.text.keywords.ops import KeyWordOps

        ops = KeyWordOps(
            credentials=self.credentials,
            fields=fields,
            model_name=model_name,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            output_fields=output_fields,
            stop_words=stop_words,
            max_keywords=max_keywords,
            nr_candidates=nr_candidates,
            use_maxsum=use_maxsum,
            use_mmr=use_mmr,
            diversity=diversity,
        )
        ops.run(
            self,
            batched=batched,
            chunksize=chunksize,
            filters=filters,
            select_fields=fields,
            output_fields=output_fields,
            **kwargs,
        )
        return ops

    def deduplicate(
        self, fields, amount_to_deduplicate: int = 100, filters: list = None
    ):
        """
        You can deduplicate values in your dataset here.

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds.deduplicate("text_field")

        """
        results = self.aggregate(
            aggregation_query=dict(
                groupby=[
                    {
                        "field": field,
                        "agg": "category",
                        "name": field,
                        "group_size": amount_to_deduplicate,
                        "select_fields": ["_id"],
                    }
                    for field in fields
                ]
            ),
            filters=filters,
            page_size=amount_to_deduplicate,
        )

        for r in tqdm(results["results"]):
            all_ids = [d["_id"] for d in r["documents"]]
            self.datasets.documents.bulk_delete(self.dataset_id, ids=all_ids[1:])
        print("Finished deduplicating!")

    def extract_nouns(
        self,
        fields: list,
        output_fields: list,
        model_name: str = "flair/chunk-english",
        cutoff_probability: float = 0.7,
        stopwords: list = None,
        filters: list = None,
        refresh: bool = False,
        chunksize: int = 50,
        **kwargs,
    ):
        """
        Extract nouns to build a taxonomy
        """
        # TODO: add support for noun processing hooks

        from relevanceai.operations_new.processing.text.extract_nouns.ops import (
            ExtractNounsOps,
        )

        ops = ExtractNounsOps(
            credentials=self.credentials,
            fields=fields,
            model_name=model_name,
            output_fields=output_fields,
            cutoff_probability=cutoff_probability,
            stopwords=stopwords,
        )

        ops.run(
            self,
            batched=True,
            chunksize=chunksize,
            filters=filters,
            select_fields=fields,
            output_fields=output_fields,
            refresh=refresh,
        )

        return ops

    def view_workflow_history(self):
        """
        View all previous workflows

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset('sample')
            ds.view_workflow_history()

        """
        metadata = self.metadata["_operationhistory_"]
        op_docs = []
        for k, v in metadata.items():
            v["time"] = k
            op_docs.append(v)

        df = pd.DataFrame(op_docs)
        df["time"] = df["time"].apply(
            lambda x: datetime.fromtimestamp(float(x.replace("-", ".")))
        )
        return df

    def translate(
        self,
        fields: list,
        model_id: str = None,
        output_fields: list = None,
        chunksize: int = 20,
        filters: list = None,
        refresh: bool = False,
    ):
        if model_id is None:
            model_id = "facebook/mbart-large-50-many-to-many-mmt"
        from relevanceai.operations_new.processing.text.translate.ops import (
            TranslateOps,
        )

        ops = TranslateOps(
            credentials=self.credentials,
            fields=fields,
            model_id=model_id,
            output_fields=output_fields,
        )

        ops.run(
            self,
            batched=True,
            chunksize=chunksize,
            filters=filters,
            select_fields=fields,
            output_fields=output_fields,
            refresh=refresh,
        )

        return ops

    def extract_ner(
        self,
        fields: list,
        model_id: str = None,
        output_fields: list = None,
        chunksize: int = 20,
        filters: list = None,
        refresh: bool = False,
    ):
        """
        Extract NER
        """
        from relevanceai.operations_new.processing.text.ner.ops import ExtractNEROps

        if model_id is None:
            model_id = "dslim/bert-base-NER"
        ops = ExtractNEROps(
            credentials=self.credentials,
            fields=fields,
            model_id=model_id,
            output_fields=output_fields,
        )

        ops.run(
            self,
            batched=True,
            chunksize=chunksize,
            filters=filters,
            select_fields=fields,
            output_fields=output_fields,
            refresh=refresh,
        )

        return ops

    def tag_text(
        self,
        fields: list,
        model_id: str = None,
        labels: list = None,
        output_fields: list = None,
        chunksize: int = 20,
        minimum_score: float = 0.2,
        maximum_number_of_labels: int = 5,
        filters: list = None,
        refresh: bool = False,
        **kwargs,
    ):
        """
        Tag Text
        """
        from relevanceai.operations_new.text_tagging.ops import TextTagOps

        ops = TextTagOps(
            credentials=self.credentials,
            fields=fields,
            model_id=model_id,
            output_fields=output_fields,
            minimum_score=minimum_score,
            maximum_number_of_labels=maximum_number_of_labels,
            labels=labels,
        )

        ops.run(
            self,
            batched=True,
            chunksize=chunksize,
            filters=filters,
            select_fields=fields,
            output_fields=output_fields,
            refresh=refresh,
            **kwargs,
        )

        return ops



from collections import defaultdict
from typing import Any, List, Optional, Union

try:
    import torch

    from torch import nn
except:
    pass


class FeatureExtractor(nn.Module):
    def __init__(self, model: nn.Module, layer_name: Optional[str] = None):
        """`__init__` is a special method in Python classes. It is the method that is called when you
        create an instance of the class

        Parameters
        ----------
        model : nn.Module
            The model to be visualized.
        layer_names : Optional[List[str]]
            A list of layer names to use for the hook. If None, the second last layer is used.

        """

        super().__init__()
        self.model = model

        layer_dict = dict([*self.model.named_modules()])

        if layer_name is None:
            layer_name = list(layer_dict.keys())[-2]

        self.layer_name = layer_name
        self._features: defaultdict = defaultdict(list)

        layer = layer_dict[layer_name]
        layer.register_forward_hook(self.save_outputs_hook(layer_name))

    def save_outputs_hook(self, layer_name):
        """It returns a function that takes 3 arguments (inputs, outputs, and a layer name) and assigns the
        output to the layer name in the _features dictionary

        Parameters
        ----------
        layer_name
            The name of the layer to save the output of.

        Returns
        -------
            A function that takes 3 arguments.

        """

        def fn(_, __, output):
            self._features[layer_name] = output

        return fn

    def forward(self, x: Union[torch.Tensor, List[Any]]) -> List[Any]:
        """The function takes in an input tensor x and passes it through the model. The output of the model
        is not used, but the output of the model's features is returned

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            The features of the model

        """
        if isinstance(x, list):
            x = torch.tensor(x)

        bs = x.shape[0]
        _ = self.model(x)
        x = self._features[self.layer_name]
        x = x.view(bs, -1)
        return x.tolist()


class EmbeddingsExtractor(nn.Module):
    def __init__(self, model: nn.Module):
        """This function takes a model as input and returns a dictionary of all the embedding layers in the
        model

        Parameters
        ----------
        model : nn.Module
            nn.Module

        """
        super().__init__()
        self.model = model

        layer_dict = dict([*self.model.named_modules()])
        self.embedding_layers = {
            k: v for k, v in layer_dict.items() if isinstance(v, nn.Embedding)
        }

    def forward(
        self,
        tokens: Union[torch.Tensor, List[int]],
        embedding_layer: Optional[str] = None,
    ) -> List[List[float]]:
        """It takes a list of tokens, and returns a list of embeddings for those tokens

        Parameters
        ----------
        tokens : List[int]
            List[int]
        embedding_layer : Optional[str]
            The name of the embedding layer to use.

        Returns
        -------
            A list of lists of floats.

        """
        if isinstance(tokens, list):
            tokens = torch.tensor(tokens).view(-1)

        if embedding_layer is None:
            embedding_layer = list(self.embedding_layers.keys())[0]

        embeddings = self.embedding_layers[embedding_layer]
        embeddings = embeddings(tokens)

        return embeddings.tolist()



from copy import deepcopy
from abc import abstractmethod

from typing import List, Dict, Any

from relevanceai.operations_new.vectorize.models.base import VectorizeModelBase
from relevanceai.operations_new.transform_base import TransformBase


class VectorizeTransform(TransformBase):

    models: List[VectorizeModelBase]
    fields: List[str]

    def __init__(
        self,
        fields: List[str],
        models: List[VectorizeModelBase],
        output_fields: list = None,
        **kwargs,
    ):
        self.fields = fields
        self.models = [self._get_model(model) for model in models]
        self.vector_fields = []
        for model in self.models:
            for field in self.fields:
                self.vector_fields.append(model.vector_name(field))
        if output_fields is not None:
            if len(output_fields) != len(models) * len(fields):
                raise NotImplementedError(
                    "Output fields only supported for 1 model for now."
                )

        self.output_fields = (
            output_fields
            if output_fields is not None
            else [self._get_output_field(t, m) for t in fields for m in models]
        )

        for key, value in kwargs.items():
            setattr(self, key, value)

    @staticmethod
    def _get_output_field(t, m):
        return f"{t}_{m}_vector_"

    @property
    def name(self):
        return "vectorizing"

    @abstractmethod
    def _get_model(self, *args, **kwargs):
        raise NotImplementedError

    def transform(self, documents: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """It takes a list of documents, and for each document, it runs the document through each of the
        models in the pipeline, and returns the updated documents.

        Parameters
        ----------
        documents : List[Dict[str, Any]]
            List[Dict[str, Any]]

        Returns
        -------
            A list of dictionaries.

        """

        updated_documents = deepcopy(documents)

        for model in self.models:
            updated_documents = model.encode_documents(
                documents=updated_documents,
                fields=self.fields,
                output_fields=self.output_fields,
            )

        # removes unnecessary info for updated_where
        updated_documents = [
            {
                key: value
                for key, value in document.items()
                if key not in self.fields or key == "_id"
            }
            for document in updated_documents
        ]

        return updated_documents

    def _get_base_filters(
        self,
    ) -> List:

        """
        Creates the filters necessary to search all documents
        within a dataset that contain fields specified in "fields"
        but do not contain their resepctive vector_fields defined in "vector_fields"

        e.g.
        fields = ["text", "title"]
        vector_fields = ["text_use_vector_", "title_use_vector_"]

        we want to search the dataset where:
        ("text" * ! "text_use_vector_") + ("title" * ! "title_use_vector_")

        Since the current implementation of filtering only accounts for CNF and not DNF boolean logic,
        We must use boolean algebra here to obtain the CNF from a DNF expression.

        CNF = Conjunctive Normal Form (Sum of Products)
        DNF = Disjunctive Normal Form (Product of Sums)

        This means converting the above to:
        ("text" + "title") * ("text" + ! "title_use_vector_") *
        (! "text_use_vector_" + "title") * (! "text_use_vector_" + ! "title_use_vector_")

        Arguments:
            fields: List[str]
                A list of fields within the dataset

            vector_fields: List[str]
                A list of vector_fields, created from the fields given the current encoders.
                These would be present if the fields in "fields" were vectorized

        Returns:
            filters: List[Dict[str, Any]]
                A list of filters.
        """

        if len(self.fields) > 1:
            iters = len(self.fields) ** 2

            filters: list = []
            for i in range(iters):
                binary_array = [character for character in str(bin(i))][2:]
                mixed_mask = ["0"] * (
                    len(self.fields) - len(binary_array)
                ) + binary_array
                mask = [int(value) for value in mixed_mask]
                # Creates a binary mask the length of fields provided
                # for two fields, we need 4 iters, going over [(0, 0), (1, 0), (0, 1), (1, 1)]

                condition_value = [
                    {
                        "field": field if mask[index] else vector_field,
                        "filter_type": "exists",
                        "condition": "==" if mask[index] else "!=",
                        "condition_value": "",
                    }
                    for index, (field, vector_field) in enumerate(
                        zip(self.fields, self.vector_fields)
                    )
                ]
                filters += [{"filter_type": "or", "condition_value": condition_value}]

        else:  # Special Case when only 1 field is provided
            condition_value = [
                {
                    "field": self.fields[0],
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": " ",
                },
                {
                    "field": self.vector_fields[0],
                    "filter_type": "exists",
                    "condition": "!=",
                    "condition_value": " ",
                },
            ]
            filters = condition_value

        return filters



from relevanceai.apps.report_app.blocks import ReportBlocks

class ConnectedChartsBlocks(ReportBlocks):
    def connected_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        chart_mode: str = "column",
        timeseries_field: str = None,
        date_interval: str = "monthly",
        add:bool=True
    ):
        sort_by = ""
        sort_direction = "Descending"
        if isinstance(sort, str):
            sort_by = self.dataset._return_sort_in_metrics(sort)
        elif isinstance(sort, dict):
            sort_by = self.dataset._return_sort_in_metrics(list(sort.keys())[0])
            sort_direction = self.dataset._return_sort_in_metrics(list(sort.values())[0])
            if sort_direction == "desc":
                sort_direction = "Descending"
            elif sort_direction == "asc":
                sort_direction = "Ascending"
            elif sort_direction in ["Descending", "Ascending"]:
                pass
            else:
                raise Exception("'sort' dictionary value needs to be of either, 'Ascending' or 'Descending'.")

        groupby = [] if groupby is None else groupby
        metrics = [] if metrics is None else metrics
        
        main_groupby = self.dataset._clean_groupby(groupby)[0]
        main_metrics = self.dataset._clean_metrics(metrics)[0]

        if chart_mode not in ["column", "bar", "line", "scatter", "timeseries"]:
            raise Exception(f"{chart_mode} not supported")

        if chart_mode == "scatter":
            if len(main_metrics) == 1:
                show_frequency = True
            elif len(main_metrics) != 2:
                raise Exception(f"{chart_mode} needs to have 2 metrics specified.")
        elif chart_mode == "timeseries":
            if not timeseries_field:
                raise Exception(f"timeseries_field cannot be unspecified for timeseries chart")

        timeseries_attrs = {}
        if timeseries_field:
            timeseries_attrs = {"field": timeseries_field, "date_interval": date_interval}            


        aggregates = [{**g, "aggType":"groupby"} for g in main_groupby] + [{**m, "aggType":"metric"} for m in main_metrics]
        chart_attrs = {
            "aggregates": aggregates,
            "dataset_id" : self.dataset.dataset_id,
            "displayType" : chart_mode,
            "sortDirection" : sort_direction,
            "title" : title,
            "showFrequencies" : show_frequency,
            "pageSize" : page_size,
            "timeseries" : timeseries_attrs
        }
        chart_attrs["sortBy"] = sort_by
        block = {
            "type": "appBlock",
            # "attrs" : {"id": str(uuid.uuid4())},
            "content" : [{
                "type" : "datasetAggregation",
                "attrs" : chart_attrs
            }]
        }
        if add:
            self.contents.append(block)
        return block

    def connected_column_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        add:bool=True
    ):
        return self.connected_chart(
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            title=title,
            page_size=page_size,
            show_frequency=show_frequency,
            chart_mode="column",
            add=add
        )

    def connected_bar_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        add:bool=True
    ):
        return self.connected_chart(
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            title=title,
            page_size=page_size,
            show_frequency=show_frequency,
            chart_mode="bar",
            add=add
        )

    def connected_line_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        add:bool=True
    ):
        return self.connected_chart(
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            title=title,
            page_size=page_size,
            show_frequency=show_frequency,
            chart_mode="line",
            add=add
        )

    def connected_scatter_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        add:bool=True
    ):
        return self.connected_chart(
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            title=title,
            page_size=page_size,
            show_frequency=show_frequency,
            chart_mode="scatter",
            add=add
        )

    def connected_timeseries_chart(
        self, 
        groupby=None, 
        metrics=None, 
        sort=None, 
        title:str="",
        page_size:int=20,
        show_frequency: bool = True,
        add:bool=True
    ):
        return self.connected_chart(
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            title=title,
            page_size=page_size,
            show_frequency=show_frequency,
            chart_mode="timeseries",
            add=add
        )


from abc import ABC, abstractmethod
from typing import Any, Dict, List

import numpy as np

from relevanceai.utils import DocUtils


class _ModelUtils(DocUtils):
    _centroids = None

    def predict_documents(self, vector_fields, documents):
        if len(vector_fields) == 1:
            vectors = self.get_field_across_documents(vector_fields[0], documents)
            return self.predict(vectors)
        raise NotImplementedError(
            "support for multiple vector fields not available right now."
        )

    def fit_predict_documents(
        self,
        vector_fields: List[str],
        documents: List[Dict[str, Any]],
        warm_start: bool = False,
    ):
        if warm_start:
            raise NotImplementedError("`warm_start` is not supported.")

        if len(vector_fields) == 1:
            vectors = self.get_field_across_documents(vector_fields[0], documents)
            cluster_labels = self.fit_predict(vectors)
            return self.format_cluster_labels(cluster_labels)
        raise NotImplementedError(
            "support for multiple vector fields not available right now."
        )

    def __str__(self):
        if hasattr(self.model):
            return str(self.model)
        else:
            return "generic_cluster_model"

    def format_cluster_label(self, label):
        """> If the label is an integer, return a string that says "cluster_" and the integer. If the label is
        a string, return the string. If the label is neither, raise an error

        Parameters
        ----------
        label
            the label of the cluster. This can be a string or an integer.

        Returns
        -------
            A list of lists.

        """
        if isinstance(label, str):
            return label
        return "cluster_" + str(label)

    def format_cluster_labels(self, labels):
        return [self.format_cluster_label(label) for label in labels]


class ClusterModelBase(ABC, _ModelUtils):
    def __init__(self, *args, **kwargs):
        model_kwargs = kwargs.pop("model_kwargs", {})
        for key, value in model_kwargs.items():
            setattr(self, key, value)

    @staticmethod
    def import_from_string(name):
        """It takes a string, splits it on the period, and then imports the module

        Parameters
        ----------
        name
            The name of the class to import.

        Returns
        -------
            The module object.

        """
        components = name.split(".")
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    @property
    def name(self):
        """It returns the name of the model class in lowercase

        Returns
        -------
            The name of the model.

        """
        return type(self.model).__name__.lower()

    def fit(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def get_centroids_from_model(self):
        if hasattr(self.model, "cluster_centers_"):
            return self.model.cluster_centers_
        elif hasattr(self.model, "get_centers"):
            return self.model.get_centers()
        else:
            return None

    def calculate_centroids(
        self, labels: np.ndarray, vectors: np.ndarray
    ) -> np.ndarray:
        centroids = self.get_centroids_from_model()
        if centroids is None:
            if not isinstance(vectors, np.ndarray):
                vectors = np.array(vectors)

            centroids = []
            for label in sorted(np.unique(labels).tolist()):
                centroid = vectors[labels == label].mean(axis=0)
                centroids.append(centroid)
            centroids = np.array(centroids)
        self._centroids = centroids
        return centroids

    @abstractmethod
    def predict(self, *args, **kwargs) -> List[Dict[str, Any]]:
        # Returns output cluster labels
        # assumes model has been trained
        # You need to run `fit` or `fit_predict` first
        # self.model.predict()
        raise NotImplementedError

    @abstractmethod
    def fit_predict(self, *args, **kwargs) -> List[int]:
        # Trains on vectors and returns output cluster labels
        # Sklearn gives some optimization between fit and predict step (idk what though)
        raise NotImplementedError



import numpy as np


def do_nothing(documents):
    return documents


def add_new_column(documents, column, value):
    for document in documents:
        document[column] = value
    return documents


def cause_error(documents):
    for d in documents:
        d["value"] = np.nan
    return documents


def cause_some_error(documents):
    MAX_ERRORS = 5
    ERROR_COUNT = 0
    for d in documents:
        if ERROR_COUNT < MAX_ERRORS:
            d["value"] = np.nan
            ERROR_COUNT += 1
    return documents



from relevanceai.client.helpers import Credentials
from relevanceai.utils.decorators.analytics import track


class Operators:
    credentials: Credentials

    @track
    def ClusterOps(
        self,
        dataset_id: str,
        vector_fields: list,
        alias: str,
        model=None,
        **kwargs,
    ):
        from relevanceai.operations_new.cluster.ops import ClusterOps

        return ClusterOps(
            credentials=self.credentials,
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            alias=alias,
            model=model,
            **kwargs,
        )

    @track
    def SubClusterOps(
        self,
        credentials,
        alias,
        dataset,
        model,
        vector_fields: list,
        parent_field: str,
    ):
        """
        Sub Cluster Ops.
        """
        from relevanceai.operations.cluster.sub import SubClusterOps

        return SubClusterOps(
            credentials=self.credentials,
            alias=alias,
            dataset=dataset,
            model=model,
            vector_fields=vector_fields,
            parent_field=parent_field,
        )

    @track
    def LabelOps(
        self,
        **kwargs,
    ):
        from relevanceai.operations_new.label.ops import LabelOps

        return LabelOps(
            credentials=self.credentials,
            **kwargs,
        )

    @track
    def VectorizeOps(
        self,
        **kwargs,
    ):
        from relevanceai.operations_new.vectorize.ops import VectorizeOps

        return VectorizeOps(
            credentials=self.credentials,
            **kwargs,
        )

    @track
    def DimReductionOps(
        self,
        **kwargs,
    ):
        from relevanceai.operations_new.vectorize.ops import VectorizeOps

        return VectorizeOps(
            credentials=self.credentials,
            **kwargs,
        )



import io
import numpy as np
from relevanceai.apps.report_app.blocks import ReportBlocks


class PlotlyReportBlock(ReportBlocks):
    """
    Adding a plotly block to Report
    """

    def plotly(
        self,
        fig,
        title: str = "",
        static: bool = False,
        width: int = None,
        height: int = None,
        add: bool = True,
        width_percentage: int = 100,
        options=None,
        **kwargs
    ):
        if options is None:
            options = {"displayLogo":False}
        try:
            import plotly
        except ImportError:
            raise ImportError(
                ".plotly requires plotly to be installed, install with 'pip install -U plotly'."
            )
        if static:
            try:
                import kaleido
            except ImportError:
                raise ImportError(
                    ".plotly 'static=True' requires kaleido to be installed, install with 'pip install -U kaleido'."
                )
            fig_image = fig.to_image(format="png", width=width, height=height)
            self.image(
                fig_image, title=title, width_percentage=width_percentage, add=add
            )
        else:
            layout = fig._layout
            if width:
                layout['width'] = width
            else:
                if "width" in layout and layout['width'] == np.inf:
                    layout['width'] = "auto"
            if height:
                layout['height'] = height
            else:
                if "height" in layout and layout['height'] == np.inf:
                    layout['height'] = "auto"
            block = {
                "type": "appBlock",
                # "attrs" : {"id": str(uuid.uuid4())},
                "content": [{
                    'attrs': {
                        'height': 'auto',
                        'data': fig._data,
                        'layout': layout,
                        'options': options,
                        'title': title,
                        'width': f'{width_percentage}%'
                    },
                    'type': 'plotlyChart'
                }]
            }
        if add:
            self.contents.append(block)
        return block


    def plotly_dendrogram(
        self,
        data,
        labels=None,
        hierarchy_method="single",
        metric="euclidean",
        orientation="right",
        color_threshold=0.75,
        **kwargs
    ):
        """
        Convenience function to plot a dendrogram.
        """
        from scipy.cluster import hierarchy

        linkage = hierarchy.linkage(data, method=hierarchy_method, metric=metric)
        try:
            import plotly.figure_factory as ff

            fig = ff.create_dendrogram(
                data,
                labels=labels,
                orientation=orientation,
                color_threshold=color_threshold,
                linkagefun=lambda x: linkage,
            )
            return self.plotly(fig, **kwargs)
        except ImportError:
            raise ImportError(
                "This requires plotly installed, install with `pip install -U plotly`"
            )
        except TypeError as e:
            raise TypeError(
                e
                + " This is a common error that can be fixed with `pip install pyyaml==5.4.1`"
            )


from typing import Optional

from relevanceai.operations_new.apibase import OperationAPIBase
from relevanceai.operations_new.scaling.base import ScalerBase


class ScaleOps(ScalerBase, OperationAPIBase):
    """
    API related Functionality for Operation
    """

    def __init__(self, alias, model, model_kwargs: Optional[dict] = None, **kwargs):
        if model_kwargs is None:
            model_kwargs = {}

        super().__init__(
            model=model,
            alias=alias,
            model_kwargs=model_kwargs,
            **kwargs,
        )
        self.model = self._get_model(
            model=model,
            alias=alias,
            model_kwargs=model_kwargs,
        )



import warnings
from relevanceai.operations_new.cluster.sub.transform import SubClusterTransform
from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.cluster.ops import ClusterOps
from typing import Optional, Union
from copy import deepcopy


class SubClusterOps(SubClusterTransform, ClusterOps):
    def __init__(
        self,
        model,
        alias: str,
        vector_fields,
        parent_field,
        dataset_id,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        min_parent_cluster_size: int = 0,
        model_kwargs: Optional[dict] = None,
        cluster_field: str = "_cluster_",
        outlier_value: Union[int, str] = -1,
        **kw,
    ):
        self.model = model
        self.alias = alias
        if len(vector_fields) > 1:
            raise NotImplementedError
        self.vector_fields = vector_fields
        self.parent_field = parent_field
        self.filters = filters if filters is not None else []
        self.cluster_ids = cluster_ids
        self.min_parent_cluster_size = min_parent_cluster_size
        self.cluster_field = cluster_field
        self.model = self._get_model(model=model, model_kwargs=model_kwargs)
        self.outlier_value = outlier_value
        self.dataset_id = dataset_id
        self.include_cluster_report = False
        for k, v in kw.items():
            setattr(self, k, v)

    def _list_unique_values(self, field):
        # TODO: Switch to a more complete subclustering approach
        facets = self.datasets.facets(
            dataset_id=self.dataset_id, fields=[field], page_size=9999
        )
        return [
            x["value"]
            for x in facets["results"][self.parent_field]
            if x["frequency"] > self.min_parent_cluster_size
        ]

    def run(self, *args, **kwargs):
        # Loop through unique cluster values first
        # then run through it
        if kwargs.get("filters") is not None:
            self.filters = kwargs.pop("filters")
        cluster_ids = (
            self._list_unique_values(field=self.parent_field)
            if self.cluster_ids is None
            else self.cluster_ids
        )
        for cluster_id in cluster_ids:
            self.parent_label = cluster_id
            print(f"Subclustering on {cluster_id}...")
            # Create filters to loop through
            new_filters = deepcopy(self.filters) + [
                {
                    "field": self.parent_field,
                    "filter_type": "exact_match",
                    "condition": "==",
                    "condition_value": cluster_id,
                }
            ]
            super().run(filters=new_filters, *args, **kwargs)

        subcluster_field_name = self._get_cluster_field_name()
        # Store the relevant metadata
        self.store_subcluster_metadata(
            parent_field=self.parent_field, cluster_field=subcluster_field_name
        )

    def store_subcluster_metadata(self, parent_field: str, cluster_field: str):
        """
        Store subcluster metadata
        """
        return self.append_metadata_list(
            field="_subcluster_",
            value_to_append={
                "parent_field": parent_field,
                "cluster_field": cluster_field,
            },
            only_unique=True,
        )

    def get_centroid_documents(self):
        centroid_vectors = {}
        if hasattr(self.model, "_centroids") and self.model._centroids is not None:
            centroid_vectors = self.model._centroids
            # get the cluster label function
            labels = range(len(centroid_vectors))
            cluster_ids = self.format_subcluster_labels(
                labels, [self.parent_label for _ in range(len(labels))]
            )
            if len(self.vector_fields) > 1:
                warnings.warn(
                    "Currently do not support inserting centroids with multiple vector fields"
                )
            centroids = [
                {"_id": k, self.vector_fields[0]: v}
                for k, v in zip(cluster_ids, centroid_vectors)
            ]
        else:
            centroids = self.create_subcluster_centroids()
        return centroids

    def create_subcluster_centroids(self):
        if len(self.vector_fields) > 1:
            raise NotImplementedError(
                "Do not currently support multiple vector fields for centroid creation."
            )

        # calculate the centroids
        centroid_vectors = self.calculate_centroids()
        return centroid_vectors

    def format_subcluster_labels(self, labels: list, parent_labels: str):
        return [
            self.format_subcluster_label(label, parent_labels[i])
            for i, label in enumerate(labels)
        ]

    def format_subcluster_label(self, label: str, parent_label: str):
        if isinstance(label, str):
            return label
        return parent_label + "-" + str(label)



"""

    Add emotion.

"""
import numpy as np
import csv
from typing import Optional
from urllib.request import urlopen
from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations_new.transform_base import TransformBase


class EmotionTransform(TransformBase):
    def __init__(
        self,
        text_fields: list,
        model_name="Emanuel/bertweet-emotion-base",
        output_fields: list = None,
        min_score: float = 0.3,
        device: int = None,
        **kwargs,
    ):
        """
        Sentiment Ops.

        Parameters
        -------------

        model_name: str
            The name of the model

        """
        self.model_name = model_name
        self.text_fields = text_fields
        self.output_fields = output_fields
        self.min_score = min_score

        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def classifier(self):
        if not hasattr(self, "device"):
            self.device = self.get_transformers_device(None)
        if not hasattr(self, "_classifier"):
            import transformers

            self._classifier = transformers.pipeline(
                "sentiment-analysis",
                return_all_scores=False,
                model=self.model_name,
                device=self.device,
            )
        return self._classifier

    def analyze_emotion(
        self,
        text: str,
    ):
        """
        Analyze sentiment
        """
        if text is None:
            return {}
        # we force limiting string because Huggingface has unreliable truncation
        output = self.classifier(text[:300], truncation=True, max_length=512)
        # [{'label': 'desire', 'score': 0.30693167448043823}]
        if output[0]["score"] > self.min_score:
            return output
        return {}

    @property
    def name(self):
        return "emotion"

    def _get_output_field(self, text_field):
        norm_name = self.model_name.replace("/", "-")
        return f"_emotion_.{text_field}.{norm_name}"

    def transform(self, documents):
        # For each document, update the field
        sentiment_docs = [{"_id": d["_id"]} for d in documents]
        for i, t in enumerate(self.text_fields):
            if self.output_fields is not None:
                output_field = self.output_fields[i]
            else:
                output_field = self._get_output_field(t)
            sentiments = [
                self.analyze_emotion(
                    self.get_field(t, doc, missing_treatment="return_empty_string"),
                )
                for doc in documents
            ]
            self.set_field_across_documents(output_field, sentiments, sentiment_docs)
        return sentiment_docs



"""
Subclustering operation
"""
import numpy as np
from tqdm.auto import tqdm
from typing import Optional, List, Union
from relevanceai.operations_new.cluster.transform import ClusterTransform


class SubClusterTransform(ClusterTransform):
    def __init__(
        self,
        model,
        alias: str,
        vector_fields,
        parent_field,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        min_parent_cluster_size: Optional[int] = None,
        model_kwargs: Optional[dict] = None,
        cluster_field: str = "_cluster_",
        outlier_value: Union[int, str] = -1,
        outlier_label: str = "outlier",
        **kw
    ):
        self.model = model
        self.alias = alias
        if len(vector_fields) > 1:
            raise NotImplementedError
        self.vector_fields = vector_fields
        self.parent_field = parent_field
        self.filters = filters
        self.cluster_ids = cluster_ids
        self.min_parent_cluster_size = min_parent_cluster_size
        self.cluster_field = cluster_field
        self.model = self._get_model(model=model, model_kwargs=model_kwargs)
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label
        for k, v in kw.items():
            setattr(self, k, v)

    def transform(self, documents):
        if not self.is_field_across_documents(self.vector_fields[0], documents):
            raise ValueError(
                "You have missing vectors in your document. You will want to apply a filter for vector fields. See here for a page of filter options: https://relevanceai.readthedocs.io/en/development/core/filters/exists.html#exists."
            )
        parent_values = self.get_field_across_documents(self.parent_field, documents)
        vectors = self.get_field_across_documents(self.vector_fields[0], documents)
        labels = self.model.fit_predict(vectors)
        sub_labels = self._format_sub_labels(parent_values, labels)
        # Get the cluster field name
        cluster_field_name = self._get_cluster_field_name()

        documents_to_upsert = [{"_id": d["_id"]} for d in documents]

        self.set_field_across_documents(
            cluster_field_name, sub_labels, documents_to_upsert
        )
        return documents_to_upsert

    def _format_sub_labels(self, parent_values: list, labels: np.ndarray) -> List[str]:
        if len(parent_values) != len(labels):
            raise ValueError("Improper logic for parent values")

        if isinstance(labels, np.ndarray):
            labels = labels.flatten().tolist()

        cluster_labels = [
            label + "-" + str(labels[i])
            if label != self.outlier_value
            else self.outlier_label
            for i, label in enumerate(parent_values)
        ]
        return cluster_labels

    def store_subcluster_metadata(self, parent_field: str, cluster_field: str):
        """
        Store subcluster metadata
        """
        return self.append_metadata_list(
            field="_subcluster_",
            value_to_append={
                "parent_field": parent_field,
                "cluster_field": cluster_field,
            },
            only_unique=True,
        )



from typing import Optional, Union, Dict, List, Mapping
from relevanceai.dataset.read.statistics import Statistics


class ClusterRead(Statistics):
    def list_closest_to_center(
        self,
        vector_fields: list,
        alias: str,
        cluster_ids: Optional[list] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[list] = None,
        facets: Optional[list] = None,
        include_vector: bool = False,
        cluster_properties_filters: Optional[Dict] = None,
        include_count: bool = False,
        include_facets: bool = False,
        verbose: bool = False,
    ):
        return self.datasets.cluster.centroids.list_closest_to_center(
            dataset_id=self.dataset_id,
            vector_fields=vector_fields,
            alias=alias,
            cluster_ids=cluster_ids,
            select_fields=select_fields,
            approx=approx,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            facets=facets,
            include_vector=include_vector,
            include_count=include_count,
            include_facets=include_facets,
            cluster_properties_filter=cluster_properties_filters,
            verbose=verbose,
        )



"""
Clean HTML
"""
from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.processing.text.html_clean.transform import (
    CleanTextTransform,
)


class CleanTextOps(CleanTextTransform, OperationAPIBase):
    """
    Clean text operations
    """



import numpy as np
from relevanceai.operations.dr.base import DimReductionBase
from typing import Optional, Dict, Any
from relevanceai.operations.cluster.constants import (
    DIM_REDUCTION,
    DIM_REDUCTION_DEFAULT_ARGS,
)


class TSNE(DimReductionBase):
    def fit_transform(
        self,
        vectors: np.ndarray,
        dr_args: Optional[Dict[Any, Any]] = DIM_REDUCTION_DEFAULT_ARGS["tsne"],
        dims: int = 3,
    ) -> np.ndarray:
        from sklearn.decomposition import PCA
        from sklearn.manifold import TSNE

        tsne = TSNE(n_components=dims, **dr_args)
        return tsne.fit_transform(vectors)



import copy
from typing import Dict, List, Optional, Union
from relevanceai.apps.explorer_app.base import ExplorerBase


class ExplorerApp(ExplorerBase):
    def preview_fields(self, fields: List = None, append=True):
        fields = [] if fields is None else fields
        self._append_or_replace(
            "default-preview-centroids-columns", fields, default_value=[]
        )
        self.document_card(fields=fields)

    def document_card(
        self,
        fields=None,
        primary_field: str = None,
        secondary_field: str = None,
        image_field: str = None,
        layout_mode: str = None,
        page_size: int = None,
        append: bool = True,
    ):
        fields_config = []
        multi_fields = copy.deepcopy(fields)
        if image_field:
            fields_config.append({"id": "image", "value": image_field})
            if image_field in multi_fields:
                multi_fields.remove(image_field)
        if primary_field:
            fields_config.append({"id": "primary", "value": primary_field})
            if primary_field in multi_fields:
                multi_fields.remove(primary_field)
        if secondary_field:
            fields_config.append({"id": "secondary", "value": secondary_field})
            if secondary_field in multi_fields:
                multi_fields.remove(secondary_field)
        if fields:
            fields_config.append({"id": "multiple-fields", "value": multi_fields})

        if "document-view-configuration" not in self.config:
            self.config["document-view-configuration"] = {
                "layout-template": "custom",
                "layout-template-configuration": {"fields": []},
            }
        if layout_mode:
            self.config["document-view-configuration"]["layout-mode"] = layout_mode
        elif "layout-mode" not in self.config["document-view-configuration"]:
            self.config["document-view-configuration"]["layout-mode"] = "grid"
        if page_size:
            self.config["document-view-configuration"]["page-size"] = page_size
        elif "page-size" not in self.config["document-view-configuration"]:
            self.config["document-view-configuration"]["page-size"] = 20

        if append:
            self.config["document-view-configuration"]["layout-template-configuration"][
                "fields"
            ] += fields_config
        else:
            self.config["document-view-configuration"]["layout-template-configuration"][
                "fields"
            ] = fields_config

    # Search and filter section
    def facets(self, fields: List = None, append=True):
        fields = [] if fields is None else fields
        self._append_or_replace("filtering-facets", fields, default_value=[])

    def sort(self, metrics: List = None, append=True):
        metrics = [] if metrics is None else metrics
        if "key-metrics" not in self.config:
            self.config["key-metrics"]["previewQuery"] = {"groupby": [], "metrics": []}
        if append:
            self.config["key-metrics"]["previewQuery"][
                "metrics"
            ] += self.dataset._clean_metrics(metrics, name_prefix="")[0]
        else:
            self.config["key-metrics"]["previewQuery"][
                "metrics"
            ] = self.dataset._clean_metrics(metrics, name_prefix="")[0]

    def sort_default(self, metric_name: str):
        if "key-metrics" not in self.config or not self.config["key-metrics"]:
            raise TypeError("'sort' needs to be specified to use 'sort_default'")
        if isinstance(metric_name, str):
            self.config["sort-default"] = self.dataset._return_sort_in_metrics(
                metric_name, self.config["key-metrics"]["previewQuery"]["metrics"]
            )
        else:
            raise TypeError("'sort_default' needs to be of type string.")

    def search_fields(self, fields: List = None):
        fields = [] if fields is None else fields
        self.config["text-fields-to-search"] = fields

    def vector_search_fields(self, vector_fields: List = None):
        vector_fields = [] if vector_fields is None else vector_fields
        self.config["vector-fields-to-search"] = vector_fields

    def search_min_relevance(self, min_relevance: float):
        self.config["search-minimum-relevance"] = min_relevance

    # Chart section
    def chart(
        self,
        groupby: List[Dict] = None,
        metrics: List[Dict] = None,
        sort: List[Dict] = None,
        page_size: int = None,
        chart_name: str = None,
        chart_mode: str = "column",
        show_frequency: bool = False,
        **kwargs
    ):
        groupby = [] if groupby is None else groupby
        metrics = [] if metrics is None else metrics

        main_groupby = self.dataset._clean_groupby(groupby)[0]
        main_metrics = self.dataset._clean_metrics(metrics)[0]
        chart_config = self._agg_to_chart_config(
            main_groupby,
            main_metrics,
            sort=sort,
            page_size=page_size,
            chart_name=chart_name,
            chart_mode=chart_mode,
            show_frequency=show_frequency,
            **kwargs
        )
        self._append_or_replace(
            "aggregation-charts",
            [self._agg_to_chart_config(**chart_config)],
            default_value=[],
        )

    def charts(self, charts, append=True):
        if isinstance(charts, dict):
            self._append_or_replace(
                "aggregation-charts",
                [self._agg_to_chart_config(**charts)],
                default_value=[],
            )
        elif isinstance(charts, list):
            self._append_or_replace(
                "aggregation-charts",
                [self._agg_to_chart_config(**c) for c in charts],
                default_value=[],
            )
        else:
            raise TypeError("'charts' needs to be a list or dictionary.")

    # Cluster section
    def cluster(self, alias: str, vector_field: str):
        self.config["cluster"] = {"alias": alias, "vector_field": vector_field}

    def cluster_charts(self, charts: List):
        self._append_or_replace("cluster-preview-config", charts, default_value=[])

    def label_clusters(self, cluster_labels: Dict):
        if "cluster" in self.config:
            self.dataset.label_clusters(
                cluster_labels,
                self.config["cluster"]["alias"],
                self.config["cluster"]["vector_field"],
            )
        else:
            raise Exception("Specify your cluster field first with .cluster.")



import uuid
from copy import deepcopy
from typing import Dict, List, Optional, Union
from relevanceai.dataset.write import Write
from relevanceai.constants.errors import FieldNotFoundError
from relevanceai.apps import ExplorerApp


class ManageApps(Write):
    """
    Set of core functions used to do CRUD on deployables/apps.
    Config = Universal config in Relevance AI for apps
    Config Input = The input to the SDK's create_app_config
    """

    def _app_url(
        self, dataset_id: str, project_id: str, deployable_id: str, app_type=None
    ):
        if app_type:
            return f"https://cloud.relevance.ai/dataset/{dataset_id}/deploy/{app_type}/{project_id}/{self.api_key}/{deployable_id}/{self.region}"
        else:
            return f"https://cloud.relevance.ai/dataset/{dataset_id}/deploy/explore/{project_id}/{self.api_key}/{deployable_id}/{self.region}"

    def list_apps(self, return_config: bool = False):
        # print(
        #     "Note: Deployable is the same as App. Deployables are legacy names of what we call Apps in the backend."
        # )
        if return_config:
            return [
                d
                for d in self.deployables.list()["deployables"]
                if d["dataset_id"] == self.dataset_id
            ]
        else:
            results = []
            for d in self.deployables.list()["deployables"]:
                if d["dataset_id"] == self.dataset_id:
                    result = {
                        "deployable_id": d["deployable_id"],
                    }
                    if "configuration" in d:
                        if "deployable_name" in d["configuration"]:
                            result["deployable_name"] = d["configuration"][
                                "deployable_name"
                            ]
                        if "type" in d["configuration"]:
                            result["type"] = d["configuration"]["type"]
                            result["url"] = self._app_url(
                                d["dataset_id"],
                                d["project_id"],
                                d["configuration"]["type"],
                                d["deployable_id"],
                            )
                    results.append(result)
            return results

    def create_app(self, config: Dict, verbose: bool = True):
        result = self.deployables.create(
            dataset_id=self.dataset_id, configuration=self.json_encoder(config)
        )
        if verbose:
            import click

            click.secho(" You can now view your app in your browser.")
            click.secho(
                f""" {self._app_url(
                    dataset_id=result['dataset_id'],
                    project_id=result['project_id'],
                    deployable_id=result['deployable_id'],
                    app_type=result['configuration']['type']
                )}""",
                bold=True,
            )
        return result

    def update_app(
        self,
        deployable_id: str,
        config: Dict,
        overwrite: bool = False,
        verbose: bool = True,
    ):
        status = self.deployables.update(
            deployable_id=deployable_id,
            dataset_id=self.dataset_id,
            configuration=self.json_encoder(config),
            overwrite=overwrite,
        )
        if status["status"] == "success":
            if "type" in config:
                app_type = config["type"]
            elif "type" in config["configuration"]:
                app_type = config["configuration"]["type"]
            if verbose:
                import click

                click.secho(" You can now view your app in your browser.")
                click.secho(
                    f""" {self._app_url(
                        dataset_id=self.dataset_id,
                        project_id=self.project,
                        deployable_id=deployable_id,
                        app_type=app_type
                    )}""",
                    bold=True,
                )
        else:
            print("Failed to update")
        return status

    def delete_app(self, deployable_id: str):
        return self.deployables.delete(deployable_id=deployable_id)

    def get_app(self, deployable_id: str):
        return self.deployables.get(deployable_id=deployable_id)

    def get_app_ids_by_name(self, name: str):
        ids = []
        for a in self.list_apps():
            if "configuration" in a:
                if (
                    "deployable_name" in a["configuration"]
                    and a["configuration"]["deployable_name"] == name
                ):
                    ids.append(a["deployable_id"])
        return ids

    def create_app_config(
        self,
        name: str = "App",
        default_view: str = "charts",
        preview_fields: List[str] = None,
        facets: List[str] = None,
        search_fields: List[str] = None,
        vector_search_fields: List[str] = None,
        search_min_relevance: float = None,
        sort: Union[List[str], List[dict]] = None,
        sort_default: str = None,
        sort_default_direction: str = None,
        charts: Union[List[str], List[dict]] = None,
        cluster_alias: str = None,
        cluster_vector_field: str = None,
        cluster_charts: Union[List[str], List[dict]] = None,
        charts_view_column: int = 2,
        preview_centroids_page_size: int = 2,
        **kwargs,
    ):
        """
        This is an easy wrapper for creating explorer app
        """
        eapp = ExplorerApp(
            name=name,
            dataset=self,
            default_view=default_view,
            charts_view_column=charts_view_column,
            preview_centroids_page_size=preview_centroids_page_size,
        )
        if preview_fields:
            eapp.preview_fields(fields=preview_fields)

        # Filter/Drilldown bar
        if facets:
            eapp.facets(fields=facets)
        if search_fields:  # make sure its text
            eapp.search_fields(fields=search_fields)
        if vector_search_fields:
            eapp.vector_search_fields(vector_fields=vector_search_fields)
        if search_min_relevance:
            eapp.search_min_relevance(min_relevance=search_min_relevance)
        if sort:
            eapp.sort(metrics=sort)
        if sort_default:
            eapp.sort_default(metric_name=sort_default)

        # Charts section
        if charts:
            eapp.charts(charts=charts)

        # Cluster section
        if cluster_alias and cluster_vector_field:
            eapp.cluster(alias=cluster_alias, vector_field=cluster_vector_field)
        elif (not cluster_alias and cluster_vector_field) or (
            cluster_alias and not cluster_vector_field
        ):
            raise KeyError(
                "Require both 'cluster_alias' and 'cluster_vector_field' for setting cluster explorer."
            )
        if cluster_charts:
            eapp.cluster_charts(charts=cluster_charts)
        return eapp.config



"""Explainer
"""
import numpy as np
import itertools
from typing import List, Callable
from doc_utils import DocUtils


class BaseExplainer(DocUtils):
    """
    Base class for vector
    """

    __name__ = ""

    @classmethod
    def chunk(self, lst: List, chunksize: int):
        """
        Chunk an iterable object in Python but not a pandas DataFrame.
        Args:
            lst:
                Python List
            chunksize:
                The chunk size of an object.
        Example:
            >>> documents = [{...}]
            >>> BaseExplainer.chunk(documents)
        """
        for i in range(0, len(lst), chunksize):
            yield lst[i : i + chunksize]

    def _vector_operation(self, vectors, vector_operation: str = "mean", axis=0):
        """
        Args:
            Vectors: the list of vectors to include
            vector_operation: One of ['mean', 'minus', 'sum', 'min', 'max']
            axis: The axis to which to perform the operation
        """
        if vector_operation == "mean":
            return np.mean(vectors, axis=axis).tolist()
        elif vector_operation == "minus":
            return np.subtract(vectors, axis=axis).tolist()
        elif vector_operation == "sum":
            return np.sum(vectors, axis=axis).tolist()
        elif vector_operation == "min":
            return np.min(vectors, axis=axis).tolist()
        elif vector_operation == "max":
            return np.max(vectors, axis=axis).tolist()
        else:
            return np.mean(vectors, axis=axis).tolist()

    def is_empty_vector(self, vector):
        return all([x == 1e-7 for x in vector])

    def get_default_vector_field_name(self, field, field_type="vector"):
        if field_type == "vector":
            return field + "_" + self.__name__ + "_vector_"
        elif field_type == "chunkvector":
            return field + "_" + self.__name__ + "_chunkvector_"

    def _encode_document(
        self,
        encode_fn,
        field,
        doc,
        vector_error_treatment="zero_vector",
        field_type: str = "vector",
    ):
        """Encode document"""
        vector = encode_fn(self.get_field(field, doc))
        if vector_error_treatment == "zero_vector":
            self.set_field(
                self.get_default_vector_field_name(field, field_type=field_type),
                doc,
                vector,
            )
            return
        elif vector_error_treatment == "do_not_include":
            return
        else:
            if vector is None or self.is_empty_vector(vector):
                vector = vector_error_treatment
            self.set_field(self.get_default_vector_field_name(field), doc, vector)

    def encode_documents(
        self,
        encode_fn,
        fields: list,
        documents: list,
        vector_error_treatment="zero_vector",
        field_type="vector",
    ):
        """
        Encode documents and their specific fields. Note that this runs off the
        default `encode` method. If there is a specific function that you want run, ensure
        that it is set to the encode function.

        Parameters:
            missing_treatment:
                Missing treatment can be one of ["do_not_include", "zero_vector", value].
            documents:
                The documents that are being used
            fields:
                The list of fields to be used
            field_type:
                Accepts "vector" or "chunkvector"
        """
        for f in fields:
            # Replace with case-switch in future
            [
                self._encode_document(
                    encode_fn,
                    f,
                    d,
                    vector_error_treatment=vector_error_treatment,
                    field_type=field_type,
                )
                for d in documents
                if self.is_field(f, d)
            ]
        return documents

    def encode_chunk_documents(
        self,
        chunk_field,
        fields: list,
        documents: list,
        vector_error_treatment: str = "zero_vector",
    ):
        """Encode chunk documents. Loops through every field and then every document.

        Parameters:
            chunk_field: The field for chunking
            fields: A list of fields for chunk documents
            documents: a list of documents
            vector_error_treatment: Vector Error Treatment

        Example:
            >>> chunk_docs = enc.encode_chunk_documents(chunk_field="value", fields=["text"], documents=chunk_docs)

        """
        # Replace with case-switch in future
        for field in fields:
            [
                self._encode_chunk_document(
                    chunk_field=chunk_field,
                    field=field,
                    doc=d,
                    vector_error_treatment=vector_error_treatment,
                    field_type="chunkvector",
                )
                for d in documents
                if self.is_field(chunk_field, d)
            ]
        return documents

    def bulk_encode_documents(
        self,
        fields: list,
        documents: list,
        vector_error_treatment="zero_vector",
        field_type="vector",
    ):
        """
        Encode documents and their specific fields. Note that this runs off the
        default `encode` method. If there is a specific function that you want run, ensure
        that it is set to the encode function.

        Parameters:
            missing_treatment:
                Missing treatment can be one of ["do_not_include", "zero_vector", value].
            documents:
                The documents that are being used
            fields:
                The list of fields to be used
        """
        for field in fields:
            # Replace with case-switch in future
            contained_docs = [d for d in documents if self.is_field(field, d)]
            self._bulk_encode_document(
                field,
                contained_docs,
                vector_error_treatment=vector_error_treatment,
                field_type=field_type,
            )
        return documents

    def get_combinations(self, lst, maximum_span=5):
        for i, j in itertools.combinations(range(len(lst) + 1), 2):
            if j - -i == len(lst):
                continue
            if j - i <= maximum_span:
                yield lst[i:j]

    def get_cosine_similarity(self, vector_1, vector_2):
        from scipy import spatial

        return 1 - spatial.distance.cosine(vector_1, vector_2)

    def get_word_combinations(self, sentence, maximum_span=5):
        for x in list(self.get_combinations(sentence.split(), maximum_span)):
            yield " ".join(x)
        yield ""

    def get_result(self, encode_fn, result_text, query_vector):
        return {
            "_search_score": self.get_cosine_similarity(
                query_vector, encode_fn(result_text)
            ),
            "text": result_text,
        }

    def explain_chunk(
        self,
        encode_fn: Callable,
        query_text,
        result_texts,
        maximum_span: int = 5,
    ):
        query_vector = encode_fn(query_text)
        cos_similarity = list()
        for result_text in result_texts:
            doc = self.get_result(encode_fn, result_text, query_vector)
            doc["explain_chunk"] = [
                {"text": t}
                for t in self.get_word_combinations(
                    result_text, maximum_span=maximum_span
                )
            ]
            self.encode_documents(encode_fn, ["text"], doc["explain_chunk"])
            for d in doc["explain_chunk"]:
                d["_search_score"] = self.get_cosine_similarity(
                    d[f"text_{self.__name__}_vector_"], query_vector
                )
            doc["explain_chunk"] = sorted(
                doc["explain_chunk"], key=lambda x: x["_search_score"], reverse=True
            )
            cos_similarity.append(doc.copy())
        return cos_similarity

    def explain(
        self,
        encode_fn: Callable,
        query_text,
        answer_text,
    ):

        query_vector = encode_fn(query_text)
        return self.explain_from_vector(
            encode_fn=encode_fn, query_vector=query_vector, answer_text=answer_text
        )

    def explain_from_vector(
        self,
        encode_fn: Callable,
        query_vector,
        answer_text,
    ):
        doc = self.get_result(encode_fn, answer_text, query_vector)
        doc["explain_chunk"] = [
            {"text": t} for t in self.get_word_combinations(answer_text)
        ]
        self.encode_documents(encode_fn, ["text"], doc["explain_chunk"])
        for d in doc["explain_chunk"]:
            d["_search_score"] = self.get_cosine_similarity(
                d[f"text_{self.__name__}_vector_"], query_vector
            )
        doc["explain_chunk"] = sorted(
            doc["explain_chunk"], key=lambda x: x["_search_score"], reverse=True
        )
        return doc["explain_chunk"][0]["text"]



"""API Client
"""
from typing import Any, Dict, List, Optional
import uuid
from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base

from relevanceai._api.endpoints.admin.admin import AdminClient
from relevanceai._api.endpoints.datasets.datasets import DatasetsClient
from relevanceai._api.endpoints.services.services import ServicesClient
from relevanceai._api.endpoints.deployables.deployables import DeployableClient
from relevanceai._api.endpoints.workflows.workflows import WorkflowsClient

from relevanceai.constants.errors import FieldNotFoundError

from relevanceai.utils.datasets import ExampleDatasets
from relevanceai.utils import make_id
from relevanceai.utils import DocUtils


class APIEndpointsClient(_Base, DocUtils):
    """API Client"""

    def __init__(
        self,
        credentials: Credentials,
        **kwargs,
    ):
        self.credentials = credentials
        super().__init__(
            credentials=credentials,
            **kwargs,
        )

    # To avoid the inits being missed - we set the endpoints as
    # properties here
    # we set the clients as internal variables
    # so they do not have to be initiated every time

    @property
    def datasets(self):
        if hasattr(self, "_datasets_client"):
            return self._datasets_client
        else:
            self._datasets_client = DatasetsClient(self.credentials)
        return self._datasets_client

    @property
    def services(self):
        if hasattr(self, "_services_client"):
            return self._services_client
        else:
            self._services_client = ServicesClient(self.credentials)
        return self._services_client

    @property
    def example_datasets(self):
        return ExampleDatasets()

    @property
    def admin(self):
        if hasattr(self, "_admin_client"):
            return self._admin_client
        self._admin_client = AdminClient(self.credentials)
        return self._admin_client

    @property
    def deployables(self):
        if hasattr(self, "_deployables_client"):
            self._deployables_client = DeployableClient(self.credentials)
        self._deployables_client = DeployableClient(self.credentials)
        return self._deployables_client

    @property
    def workflows(self):
        if hasattr(self, "_workflows_client"):
            self._workflows_client = WorkflowsClient(self.credentials)
        self._workflows_client = WorkflowsClient(self.credentials)
        return self._workflows_client

    def _validate_ids(self, documents: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Not inplace anymore
        """
        return [
            {"_id": str(uuid.uuid4()), **document}
            if "_id" not in document
            else document
            for document in documents
        ]

    def _are_fields_in_schema(
        self,
        fields: List[str],
        dataset_id: str,
        schema: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Check fields are in schema
        """
        if schema is None:
            schema = self.datasets.schema(dataset_id)
        invalid_fields = []
        for i in fields:
            if i not in schema:
                invalid_fields.append(i)
        if len(invalid_fields) > 0:
            raise ValueError(
                f"{', '.join(invalid_fields)} are invalid fields. They are not in the dataset schema."
            )
        return



"""Base64 decoding for workflows
"""
import os
import base64
import json


def decode_workflow_token(token):
    """It takes a token, decodes it, and returns the decoded token

    Parameters
    ----------
    token
        The token that was generated by the workflow.

    Returns
    -------
        A dictionary of the workflow configuration.

    """
    config = json.loads(base64.b64decode(token + "==="))
    # Set workflow ID for tracking
    os.environ["WORKFLOW_ID"] = config.get("job_id", "")
    return config



"""
Base class for clustering
"""
import numpy as np
import warnings
from typing import List, Dict, Any, Optional
from relevanceai.operations_new.cluster.models.base import ClusterModelBase
from relevanceai.operations_new.transform_base import TransformBase
from relevanceai.operations_new.cluster.alias import ClusterAlias


class ClusterTransform(TransformBase, ClusterAlias):

    model: ClusterModelBase

    def __init__(
        self,
        vector_fields: List[str],
        alias: str,
        model: Any,
        model_kwargs: Optional[dict] = None,
        cluster_field: str = "_cluster_",
        include_cluster_report: bool = False,
        **kwargs,
    ):

        self.vector_fields = vector_fields
        self.cluster_field = cluster_field

        self.model_kwargs = {} if model_kwargs is None else model_kwargs
        self.alias = self._get_alias(alias)
        self.model = self._get_model(model=model, model_kwargs=self.model_kwargs)

        self.include_cluster_report = include_cluster_report

        for k, v in kwargs.items():
            setattr(self, k, v)
        self._check_vector_fields()

    @property
    def name(self):
        return "cluster"

    def fit_predict_documents(self, documents, warm_start=False):
        """
        If warm_start=True, copies the values from the previous fit.
        Only works for cluster models that use centroids. You should
        not have to use this parameter.
        """
        # run fit predict on documetns
        if hasattr(self.model, "fit_predict_documents"):
            return self.model.fit_predict_documents(
                documents=documents,
                vector_fields=self.vector_fields,
                warm_start=warm_start,
            )
        elif hasattr(self.model, "fit_predict"):
            if len(self.vector_fields) == 1:
                vectors = self.get_field_across_documents(
                    self.vector_fields[0],
                    documents,
                )
                cluster_labels = self.model.fit_predict(
                    vectors,
                    # warm_start=warm_start,
                )

                return self.format_cluster_labels(cluster_labels)
        raise AttributeError("Model is missing a `fit_predict` method.")

    def transform(self, documents: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """It takes a list of documents, and for each document, it runs the document through each of the
        models in the pipeline, and returns the updated documents.

        Parameters
        ----------
        documents : List[Dict[str, Any]]
            List[Dict[str, Any]]

        Returns
        -------
            A list of dictionaries.

        """

        if not self.is_field_across_documents(self.vector_fields[0], documents):
            raise ValueError(
                "You have missing vectors in your document. You need to filter out documents that don't contain the vector field."
            )
        labels = self.fit_predict_documents(
            documents=documents,
        )
        # Get the cluster field name
        cluster_field_name = self._get_cluster_field_name()

        documents_to_upsert = [{"_id": d["_id"]} for d in documents]
        self.set_field_across_documents(cluster_field_name, labels, documents_to_upsert)
        if hasattr(self, "include_cluster_report") and self.include_cluster_report:
            vectors = self.get_field_across_documents(
                self.vector_fields[0],
                documents,
            )
            try:
                self.set_field_across_documents(
                    self._silhouette_score_field_name(),
                    self.calculate_silhouette(vectors, labels),
                    documents_to_upsert,
                )
            except:
                pass
            try:
                self.set_field_across_documents(
                    self._squared_error_field_name(),
                    self.calculate_squared_error(
                        vectors, labels, self.model._centroids
                    ),
                    documents_to_upsert,
                )
            except:
                import traceback

                traceback.print_exc()
                pass
        return documents_to_upsert

    @staticmethod
    def calculate_silhouette(vectors, labels):
        try:
            from sklearn.metrics import silhouette_samples

            return silhouette_samples(vectors, labels)
        except ImportError:
            raise ImportError("sklearn missing")
        except:
            raise Exception("Couldn't calculate silhouette scores")

    def _silhouette_score_field_name(self):
        return f"_silhouette_score_{self.alias}"

    @staticmethod
    def calculate_squared_error(vectors, labels, centroids):
        try:
            label_to_centroid_index = {
                label: i for i, label in enumerate(sorted(np.unique(labels).tolist()))
            }
            return np.square(
                np.subtract(
                    [centroids[label_to_centroid_index[l]] for l in labels],
                    vectors,
                )
            )
        except:
            raise Exception("Couldn't calculate squared errors")

    def _squared_error_field_name(self):
        return f"_squared_error_{self.alias}"

    def _get_cluster_field_name(self):
        alias = self.alias
        if isinstance(self.vector_fields, list):
            if hasattr(self, "cluster_field"):
                set_cluster_field = (
                    f"{self.cluster_field}.{'.'.join(self.vector_fields)}.{alias}"
                )
            else:
                set_cluster_field = f"_cluster_.{'.'.join(self.vector_fields)}.{alias}"

        elif isinstance(self.vector_fields, str):
            set_cluster_field = f"{self.cluster_field}.{self.vector_fields}.{alias}"

        return set_cluster_field

    def format_cluster_label(self, label):
        """> If the label is an integer, return a string that says "cluster-" and the integer. If the label is
        a string, return the string. If the label is neither, raise an error

        Parameters
        ----------
        label
            the label of the cluster. This can be a string or an integer.

        Returns
        -------
            A list of lists.

        """
        if isinstance(label, str):
            return label
        return "cluster_" + str(label)

    def format_cluster_labels(self, labels):
        return [self.format_cluster_label(label) for label in labels]



from typing import Union, Any, Optional, List

from relevanceai.utils import DocUtils

from relevanceai.client.helpers import Credentials
from relevanceai.operations import BaseOps
from relevanceai._api import APIClient


class ReduceDimensionsOps(APIClient, BaseOps):
    def __init__(
        self,
        credentials: Credentials,
        n_components: int,
        model: Any,
        vector_fields: List[str],
        dr_field: str = "_dr_",
        verbose: bool = True,
        dataset_id: Optional[str] = None,
        alias: Optional[str] = None,
    ):
        from relevanceai.operations.dr.models import PCA, TSNE, UMAP, Ivis

        if isinstance(model, str):
            algorithm = model.upper()
            if algorithm == "PCA":
                model = PCA()
            elif algorithm == "TSNE":
                model = TSNE()
            elif algorithm == "UMAP":
                model = UMAP()
            elif algorithm == "IVIS":
                model = Ivis()
            else:
                raise ValueError()

        self.model = model
        self.dr_field = dr_field
        self.verbose = verbose
        self.n_components = n_components
        self.dataset_id = dataset_id
        self.vector_fields = vector_fields
        self.alias = alias

        super().__init__(credentials)

    def _insert_metadata(
        self,
        dataset_id: Optional[str] = None,
        alias: Optional[str] = None,
        vector_fields: Optional[List[str]] = None,
    ):
        """
        {
            "alias" : alias,
            "vector_fields" : vector_fields,
            "n_components" : n_components,
            "params" : other_parameters_used
        }
        """
        if dataset_id is None:
            dataset_id = self.dataset_id
        if alias is None:
            alias = self.alias
        if vector_fields is None:
            vector_fields = self.vector_fields

        if dataset_id is not None:
            metadata = self.datasets.metadata(dataset_id=dataset_id)

            if "_dr_" not in metadata:
                metadata["_dr_"] = {}
            metadata["_dr_"][alias] = {
                "alias": alias,
                "vector_fields": vector_fields,
                "n_components": self.n_components,
            }

            self.datasets.post_metadata(dataset_id, metadata)

    def fit(
        self,
        vector_fields: List[str],
        dataset_id: Optional[str] = None,
        alias: Optional[str] = None,
        filters: Optional[list] = None,
        show_progress_bar: bool = True,
        verbose: bool = True,
    ):
        """
        Reduce Dimensions

        Example
        ---------


        Parameters
        -------------

        dataset_id: Optional[str]
            The dataset to run dimensionality reduction on
        vector_fields: Optional[List[str]]
            List of vector fields
        alias: Optional[str]
            Alias to store dimensionality reduction model
        show_progress_bar: bool
            If True, the progress bar can be shown

        """
        filters = [] if filters is None else filters
        if dataset_id is None:
            dataset_id = self.dataset_id

        if vector_fields is None:
            vector_fields = self._check_vector_fields(self.vector_fields)
        vector_fields = self._check_vector_fields(vector_fields)

        if alias is None:
            alias = self._create_alias(self.alias, vector_fields)
        alias = self._create_alias(alias, vector_fields)

        # check if alias == any vector fields
        for vector_field in vector_fields:
            if alias == vector_field:
                raise ValueError(
                    "Your alias is same as one of the vector_field. Relevance does not support overriding an existing vector field with a lower dimension vector."
                )

        print("Retrieving all documents...")

        from relevanceai.utils.filter_helper import create_filter

        for vector_field in vector_fields:
            filters += create_filter(vector_field, filter_type="exists")
        documents = self._get_all_documents(
            dataset_id=dataset_id,
            select_fields=vector_fields,
            show_progress_bar=show_progress_bar,
            include_vector=True,
            filters=filters,
        )

        print("Predicting on all documents...")
        dr_documents = self.model.fit_transform_documents(
            vector_fields=vector_fields,  # type: ignore
            documents=documents,
            alias=alias,  # type: ignore
            dims=self.n_components,
        )

        print("Updating dataset with reduced vectors...")
        self._update_documents(dataset_id=dataset_id, documents=dr_documents)  # type: ignore

        print("Updating dr metadata...")
        self._insert_metadata(
            dataset_id=dataset_id, vector_fields=vector_fields, alias=alias
        )

        if verbose:
            self._print_app_link()

    def _print_app_link(self):
        print("All complete")
        # link = CLUSTER_APP_LINK.format(self.dataset_id)
        # print(Messages.BUILD_HERE + link)

    def run(
        self,
        vector_fields: List[str],
        dataset_id: Optional[str] = None,
        alias: Optional[str] = None,
        filters: Optional[list] = None,
        show_progress_bar: bool = True,
        verbose: bool = True,
    ):
        """Run dimensionality reduction"""
        self.fit(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            alias=alias,
            filters=filters,
            show_progress_bar=show_progress_bar,
            verbose=verbose,
        )



import numpy as np

from relevanceai.utils.decorators.analytics import track


@track
def euclidean_distance_matrix(a, b, decimal=None):
    A = np.array(a)
    B = np.array(b)
    M = A.shape[0]
    N = B.shape[0]
    A_dots = (A * A).sum(axis=1).reshape((M, 1)) * np.ones((1, M))
    B_dots = (B * B).sum(axis=1).reshape((N, 1)) * np.ones((1, N))
    dist = A_dots + B_dots - 2 * A.dot(B.T)
    dist[dist < 1e-9] = 0
    dist = np.sqrt(dist + dist.T)
    if decimal:
        dist = np.around(dist, decimal)
    return dist.tolist()



models = {
    "affinitypropagation": "AffinityPropagation",
    "agglomerativeclustering": "AgglomerativeClustering",
    "birch": "Birch",
    "dbscan": "DBSCAN",
    "optics": "OPTICS",
    "kmeans": "KMeans",
    "featureagglomeration": "FeatureAgglomeration",
    "meanshift": "MeanShift",
    "minibatchkmeans": "MiniBatchKMeans",
    "spectralclustering": "SpectralClustering",
    "affinitypropagation": "affinity_propagation",
    "spectralbiclustering": "SpectralBiclustering",
    "spectralcoclustering": "SpectralCoclustering",
}



"""Transformers
"""
from relevanceai.operations_new.processing.transformers.transform import (
    TransformersPipelineTransform,
)
from relevanceai.operations_new.ops_base import OperationAPIBase


class TransformersPipelineOps(TransformersPipelineTransform, OperationAPIBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



models = {
    "binarizer": "Binarizer",
    "functiontransformer": "FunctionTransformer",
    "kbinsdiscretizer": "KBinsDiscretizer",
    "kernelcenterer": "KernelCenterer",
    "labelbinarizer": "LabelBinarizer",
    "labelencoder": "LabelEncoder",
    "multilabelbinarizer": "MultiLabelBinarizer",
    "quantiletransformer": "QuantileTransformer",
    "normalizer": "Normalizer",
    "onehotencoder": "OneHotEncoder",
    "ordinalencoder": "OrdinalEncoder",
    "powertransformer": "PowerTransformer",
    "splinetransformer": "SplineTransformer",
    "minmax": "MinMaxScaler",
    "maxabs": "MaxAbsScaler",
    "robust": "RobustScaler",
    "standard": "StandardScaler",
    "minmaxscaler": "MinMaxScaler",
    "maxabsscaler": "MaxAbsScaler",
    "robustscaler": "RobustScaler",
    "standardscaler": "StandardScaler",
}



from typing import Any, Dict, List, Tuple, Union


class ReportBase:
    def __init__(self, name: str, dataset, deployable_id: str = None):
        self.name = name
        self.dataset = dataset
        self.dataset_id = dataset.dataset_id
        self.deployable_id = deployable_id
        self.base_url = "https://cloud.relevance.ai"
        app_config = None
        self.reloaded = False
        if deployable_id:
            try:
                app_config = self.dataset.get_app(deployable_id)
                self.reloaded = True
            except:
                print(
                    f"{deployable_id} does not exist in the dataset, the given id will be used for creating a new app."
                )
            self.deployable_id = deployable_id
        if app_config:
            self.config = app_config["configuration"]
        else:
            self.config = {
                "dataset_name": self.dataset_id,
                "deployable_name": self.name,
                "type": "page",
                "page-content": {"type": "doc", "content": []},
            }
        if self.deployable_id:
            self.config["deployable_id"] = self.deployable_id

    @property
    def contents(self) -> List:
        return self.config["page-content"]["content"]

    def refresh(self, verbose=True, prompt_update: bool = False):
        try:
            app_config = self.dataset.get_app(self.deployable_id)["configuration"]
            self.reloaded = True
        except:
            raise Exception(f"{self.deployable_id} does not exist in the dataset.")
        if self.config == app_config:
            if verbose:
                print("No updates made, no differences detected.")
            return {}
        else:
            if verbose:
                print("Differences deteced app update made, returning the differences")
            self.config = app_config
            return {}

    def reset(self):
        self.config = {
            "dataset_name": self.dataset_id,
            "deployable_name": self.name,
            "type": "page",
            "page-content": {"type": "doc", "content": []},
        }

    def deploy(self, overwrite: bool = False, new: bool = False):
        if new:
            return self.dataset.create_app(
                {k: v for k, v in self.config.items() if k != "deployable_id"}
            )
        else:
            if self.deployable_id:
                status = self.dataset.update_app(
                    self.deployable_id, self.config, overwrite=overwrite
                )
                if status["status"] == "success":
                    return self.dataset.get_app(self.deployable_id)
                else:
                    raise Exception("Failed to update app")
            result = self.dataset.create_app(self.config)
            self.deployable_id = result["deployable_id"]
            self.reloaded = True
            return result

    def app_url(
        self,
    ):
        return f"{self.base_url}/dataset/{self.dataset_id}/deploy/page/{self.dataset.project}/{self.dataset.api_key}/{self.deployable_id}/{self.dataset.region}"

    def gui(self, width: int = 1000, height: int = 800):
        try:
            self.dataset.get_app(self.deployable_id)
        except:
            raise Exception(
                f"{self.deployable_id} does not exist in the dataset, run `.deploy` first to create the app."
            )
        return self._show_ipython(url=self.app_url(), width=width, height=height)

    def _show_ipython(self, url, width: int = 1000, height: int = 800):
        try:
            from IPython.display import IFrame

            return IFrame(url, width=width, height=height)
        except:
            print("This only works within an IPython, Notebook environment.")
            return url

    # def replace_by_id(self, id):
    #     """replace an existing block with another"""
    #     return

    # def generate_code(self):
    #     """generate python code from json"""
    #     return


from typing import Dict

TFHUB_MODELS: Dict[str, Dict[str, int]] = {
    "https://tfhub.dev/google/universal-sentence-encoder/4": {
        "vector_length": 512,
    },
}

SENTENCE_TRANSFORMER_MODELS: Dict[str, Dict[str, int]] = {
    "distilroberta-base-paraphrase-v1": {
        "vector_length": 768,
    },
    "xlm-r-distilroberta-base-paraphrase-v1": {
        "vector_length": 768,
    },
    "paraphrase-xlm-r-multilingual-v1": {
        "vector_length": 768,
    },
    "distilbert-base-nli-stsb-mean-tokens": {
        "vector_length": 768,
    },
    "bert-large-nli-stsb-mean-tokens": {
        "vector_length": 1024,
    },
    "roberta-base-nli-stsb-mean-tokens": {
        "vector_length": 768,
    },
    "roberta-large-nli-stsb-mean-tokens": {
        "vector_length": 1024,
    },
    "distilbert-base-nli-stsb-quora-ranking": {
        "vector_length": 768,
    },
    "distilbert-multilingual-nli-stsb-quora-ranking": {
        "vector_length": 768,
    },
    "distilroberta-base-msmarco-v1": {
        "vector_length": 768,
    },
    "distiluse-base-multilingual-cased-v2": {
        "vector_length": 512,
    },
    "xlm-r-bert-base-nli-stsb-mean-tokens": {
        "vector_length": 768,
    },
    "bert-base-wikipedia-sections-mean-tokens": {
        "vector_length": 768,
    },
    "LaBSE": {
        "vector_length": 768,
    },
    "average_word_embeddings_glove.6B.300d": {
        "vector_length": 300,
    },
    "average_word_embeddings_komninos": {
        "vector_length": 300,
    },
    "average_word_embeddings_levy_dependency": {
        "vector_length": 768,
    },
    "average_word_embeddings_glove.840B.300d": {
        "vector_length": 300,
    },
    "paraphrase-xlm-r-multilingual-v1": {
        "vector_length": 768,
    },
    "all-MiniLM-L6-v2": {
        "vector_length": 384,
    },
}



"""
Vectorizer model is used to interact with `advanced_vectorize`
"""
from relevanceai.operations.vector import Base2Vec


class Vectorizer(Base2Vec):
    def __init__(
        self,
        field: str,
        alias: str,
        model: callable = None,
        bulk_encode: callable = None,
        encode: callable = None,
    ):
        self.field = field
        self.alias = alias
        self.model = model
        self.bulk_encode = bulk_encode
        self.encode = encode



import pandas as pd
from typing import Dict, List, Any
from .errors import MissingFieldError


class DocReadUtils:
    """This is created as a Mixin for others to easily add to their classes"""

    @classmethod
    def get_field(self, field: str, doc: Dict, missing_treatment="raise_error"):
        """
        For nested dictionaries, tries to access a field.
        e.g.
        field = kfc.item
        This should return "chickens" based on doc below.
        {
            "kfc": {
                "item": "chickens"
            }
        }
        Args:
            field:
                Field of a document.
            doc:
                document
            missing_treatment:
                Can be one of return_empty_string/return_none/raise_error
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': 'chicken'}}
            >>> vi_client.get_field('kfc.item', sample_document) == 'chickens'
        """
        d = doc
        for f in field.split("."):
            try:
                d = d[f]
            except KeyError:
                try:
                    return doc[field]
                except KeyError:
                    if missing_treatment == "return_none":
                        return None
                    elif missing_treatment == "return_empty_string":
                        return ""
                    elif missing_treatment == "raise_error":
                        raise MissingFieldError(
                            "Document is missing " + f + " of " + field
                        )
                    else:
                        return missing_treatment
            except TypeError:
                if self._is_string_integer(f):
                    # Get the Get the chunk document out.
                    try:
                        d = d[int(f)]
                    except IndexError:
                        pass
                else:
                    if missing_treatment == "return_none":
                        return None
                    elif missing_treatment == "return_empty_string":
                        return ""
                    raise MissingFieldError("Document is missing " + f + " of " + field)
        return d

    @classmethod
    def _is_string_integer(cls, x):
        """Test if a string is numeric"""
        try:
            int(x)
            return True
        except:
            return False

    @classmethod
    def get_fields(
        self, fields: List[str], doc: Dict, missing_treatment="return_empty_string"
    ) -> List[Any]:
        """
        For nested dictionaries, tries to access a field.
        e.g.
        field = kfc.item
        This should return "chickens" based on doc below.
        {
            "kfc": {
                "item": "chickens"
            }
        }
        Args:
            fields:
                List of fields of a document.
            doc:
                document
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': 'chicken'}}
            >>> vi_client.get_field('kfc.item', sample_document) == 'chickens'
        """
        return [self.get_field(f, doc, missing_treatment) for f in fields]

    def get_field_across_documents(
        self,
        field: str,
        docs: List[Dict],
        missing_treatment: str = "return_empty_string",
    ):
        """
        For nested dictionaries, tries to access a field.
        e.g.
        field = kfc.item
        This should return "chickens" based on doc below.
        {
            "kfc": {
                "item": "chickens"
            }
        }
        Args:
            fields:
                List of fields of a document.
            doc:
                document
            missing_treatment:
                This can be one of 'skip', 'return_empty_string'

        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> documents = vi_client.create_sample_documents(10)
            >>> vi_client.get_field_across_documents('size.cm', documents)
            # returns 10 values in the nested dictionary
        """
        if missing_treatment == "skip":
            # Returns only the relevant documents
            def check_field_in_doc(doc):
                return self.is_field(field, doc)

            return [
                self.get_field(field, doc) for doc in filter(check_field_in_doc, docs)
            ]
        return [self.get_field(field, doc, missing_treatment) for doc in docs]

    def get_fields_across_document(
        self, fields: List[str], doc: Dict, missing_treatment="return_empty_string"
    ):
        """
        Get numerous fields across a document.
        """
        return [
            self.get_field(f, doc, missing_treatment=missing_treatment) for f in fields
        ]

    def get_fields_across_documents(
        self,
        fields: List[str],
        docs: List[Dict],
        missing_treatment="return_empty_string",
    ):
        """Get numerous fields across documents.

        Example
        ----------

        For document:

        docs = [
            {
                "value": 2,
                "type": "car"
            },
            {
                "value": 10,
                "type": "bus"
            }
        ]

        >>> DocUtils().get_fields_across_documents(["value", "type"], docs)
        >>> [2, "car", 10, "bus"]

        Parameters
        --------------

        missing_treatment: str
            Can be one of ["skip", "return_empty_string", "return_none", "skip_if_any_missing"]
            If "skip_if_any_missing", the document will not be included if any field is missing

        """
        if missing_treatment == "skip_if_any_missing":
            docs = self.filter_docs_for_fields(fields, docs)
            return [self.get_fields_across_document(fields, doc) for doc in docs]
        return [
            self.get_fields_across_document(
                fields, doc, missing_treatment=missing_treatment
            )
            for doc in docs
        ]

    def filter_docs_for_fields(self, fields: List, docs: List):
        """
        Filter for docs if they contain a list of fields
        """

        def is_any_field_missing_in_doc(doc):
            return all([self.is_field(f, doc) for f in fields])

        return filter(is_any_field_missing_in_doc, docs)

    @classmethod
    def is_field(self, field: str, doc: Dict) -> bool:
        """
        For nested dictionaries, tries to access a field.
        e.g.
        field = kfc.item
        This should return "chickens" based on doc below.
        {
            "kfc": {
                "item": "chickens"
            }
        }
        Args:
            collection_name:
                Name of collection.
            job_id:
                ID of the job.
            job_name:
                Name of the job.
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> sample_document = {'kfc': {'item': 'chicken'}}
            >>> vi_client.is_field('kfc.item', sample_document) == True
        """
        d = doc
        for f in field.split("."):
            try:
                d = d[f]
            except KeyError:
                try:
                    doc[field]
                    return True
                except KeyError:
                    try:
                        d[field]
                    except KeyError:
                        return False
            except TypeError:
                # To Support integers
                if self._is_string_integer(f):
                    # Get the Get the chunk document out.
                    try:
                        d = d[int(f)]
                    except IndexError:
                        pass
                else:
                    return False
        return True

    def is_field_across_documents(self, field, documents):
        return all([self.is_field(field, doc)] for doc in documents)

    @staticmethod
    def list_doc_fields(doc: dict) -> List[str]:
        """returns all fields in a document, nested fields are flattened
        example:
        input: doc = {'a': {'b':'v', 'c':'v'},
                      'd':'v'}
                      'e':{'f':{'g':'v'}
        output: ['d', 'a.b', 'a.c', 'e.f.g']
        """
        df = pd.json_normalize(doc, sep=".")
        return list(df.columns)

    @classmethod
    def subset_documents(
        self,
        fields: List[str],
        docs: List[Dict],
        missing_treatment: str = "return_none",
    ) -> List[Dict]:
        """
        Args:
            fields:
                A list of fields of interest.
            docs:
                A list of documents that may or may not have the chosen
                fields.
            missing_treatment:
                Cane be on of return_empty_string/return_none/raise_error
        Example:
            >>> from vectorai.client import ViClient
            >>> vi_client = ViClient(username, api_key, vectorai_url)
            >>> docs = [
            ...     {"kfc": {"food": "chicken nuggets", "drink": "soda"}}
            ...     {"mcd": {"food": "hamburger", "drink": "pop"}}
            ... ]
            >>> fields = [
            ...     "kfc.food", "kfc.drink", "mcd.food", "mcd.drink"
            ... ]
            >>> vi_client.subset_documents(fields, docs) == [
            ...     {
            ...         "kfc.food": "chicken nuggets", "kfc.drink": "soda"},
            ...         "mcd.food": "", "mcd.drink": ""
            ...     },
            ...     {
            ...         "kfc.food": "", "kfc.drink": ""},
            ...         "mcd.food": "hamburger", "mcd.drink": "pop"},
            ...     }
            ... ]
        """
        return [
            {
                field: self.get_field(field, doc, missing_treatment=missing_treatment)
                for field in fields
            }
            for doc in docs
        ]



import pytest

from relevanceai.dataset import Dataset


@pytest.mark.xfail(reason="not sure")
def test_batch_cluster(test_dataset: Dataset):

    vector_field = "sample_1_vector_"
    n_clusters = 15
    model = "minibatchkmeans"

    test_dataset.batch_cluster(
        vector_fields=[vector_field],
        model=model,
        model_kwargs={"n_clusters": n_clusters},
    )
    assert f"_cluster_.{vector_field}.{model}-{n_clusters}" in test_dataset.schema


@pytest.mark.xfail(reason="not sure")
def test_batch_cluster_kmeans_integration(test_dataset: Dataset):
    from sklearn.cluster import MiniBatchKMeans

    model = MiniBatchKMeans(n_clusters=20)
    test_dataset.batch_cluster(
        vector_fields=["sample_1_vector_"],
        model=model,
    )
    assert f"_cluster_.sample_1_vector_.minibatchkmeans-20" in test_dataset.schema



from abc import ABC, abstractmethod

from typing import Any, List, Union

from relevanceai.utils import DocUtils


class DimReductionModelBase(ABC, DocUtils):
    model_name: str
    alias: Union[str, None]

    def vector_name(self, fields: List[str], output_field: str = None) -> str:
        if output_field is not None:
            return output_field
        if isinstance(self.alias, str):
            if "_vector_" in self.alias:
                return self.alias
            else:
                return f"{self.alias}_vector_"
        else:
            if len(fields) > 1:
                return f"concat_{self.model_name}_vector_"
            else:
                return f"{self.model_name}_{fields[0]}_vector_"

    def fit(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def fit_transform(self, *args, **kwargs) -> Any:
        raise NotImplementedError



"""
Base class for base.py to inherit.
All functions related to running operations on datasets.

The Pull Transform Push library is designed to be able to consistently
data from Relevance AI Database, transform and then constantly push data 
to the Relevance AI Database. This ensures that resources are utilised 
to their limits.

"""
import sys
import time
import psutil
import threading
import multiprocessing as mp
import warnings

from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, List, Tuple, Type, Union, Optional, Callable
from relevanceai.constants.constants import CONFIG

from relevanceai.dataset import Dataset
from relevanceai.operations_new.transform_base import TransformBase

from tqdm.auto import tqdm

from relevanceai.utils.helpers.helpers import getsizeof


class PullTransformPush:

    pull_count: int
    transform_count: int
    push_count: int

    pull_bar: tqdm
    transform_bar: tqdm
    push_bar: tqdm

    pull_thread: threading.Thread
    transform_threads: List[threading.Thread]
    push_threads: List[threading.Thread]

    pull_dataset: Dataset
    push_dataset: Dataset

    def __init__(
        self,
        dataset: Optional[Dataset] = None,
        pull_dataset: Optional[Dataset] = None,
        func: Optional[Callable] = None,
        push_dataset: Optional[Dataset] = None,
        func_args: Optional[Tuple[Any, ...]] = None,
        func_kwargs: Optional[Dict[str, Any]] = None,
        multithreaded_update: bool = False,
        pull_chunksize: Optional[int] = None,
        warmup_chunksize: Optional[int] = None,
        transform_chunksize: Optional[int] = 128,
        push_chunksize: Optional[int] = None,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        transform_workers: int = 1,
        push_workers: int = 1,
        buffer_size: int = 0,
        show_progress_bar: bool = True,
        show_pull_progress_bar: bool = True,
        show_transform_progress_bar: bool = True,
        show_push_progress_bar: bool = True,
        ingest_in_background: bool = False,
        run_in_background: bool = False,
        ram_ratio: float = 0.8,
        update_all_at_once: bool = False,
        retry_count: int = 3,
        after_id: Optional[List[str]] = None,
        pull_limit: Optional[int] = None,
        timeout: Optional[int] = None,
    ):
        """
        Buffer size:
            number of documents to be held in limbo by both queues at any one time

        """
        super().__init__()

        if dataset is None:
            self.push_dataset = push_dataset  # type: ignore
            self.pull_dataset = pull_dataset  # type: ignore

        else:
            self.push_dataset = dataset
            self.pull_dataset = dataset

        if dataset is None and pull_dataset is None and push_dataset is None:
            raise ValueError(
                "Please set `dataset=` or `push_dataset=` and `pull_dataset=`"
            )

        self.pull_dataset_id = self.pull_dataset.dataset_id
        self.push_dataset_id = self.push_dataset.dataset_id

        self.config = CONFIG

        self.pull_limit = pull_limit

        self.failed_documents_count = 0

        if pull_limit is None:
            ndocs = self.pull_dataset.get_number_of_documents(
                dataset_id=self.pull_dataset_id,
                filters=filters,
            )
            self.ndocs = ndocs
        else:
            self.ndocs = pull_limit

        self.pull_chunksize = pull_chunksize
        self.transform_chunksize = min(transform_chunksize, self.ndocs)
        self.warmup_chunksize = warmup_chunksize
        self.push_chunksize = push_chunksize

        self.update_all_at_once = update_all_at_once
        if update_all_at_once:
            self.transform_chunksize = self.ndocs

        if func is None:
            tqdm.write(f"Transform Chunksize: {self.transform_chunksize:,}")

        self.ingest_in_background = ingest_in_background

        self.filters = [] if filters is None else filters
        self.select_fields = [] if select_fields is None else select_fields

        self.general_lock = threading.Lock()

        self.transform_batch_lock = threading.Lock()
        self.push_batch_lock = threading.Lock()

        self.func_lock: Union[threading.Lock, None]

        if not multithreaded_update:
            self.func_lock = threading.Lock()
            self.transform_workers = 1
        else:
            self.func_lock = None
            self.transform_workers = transform_workers

        self.push_workers = push_workers

        self.func_args = () if func_args is None else func_args
        self.func_kwargs = {} if func_kwargs is None else func_kwargs

        if update_all_at_once:
            self.single_queue_size = self.ndocs

        else:
            if buffer_size == 0:
                ram_size = psutil.virtual_memory().total  # in bytes

                # assuming documents are 1MB, this is an upper limit and accounts for alot
                average_size = 2**20
                max_document_size = min(average_size, 2**20)

                total_queue_size = int(ram_size * ram_ratio / max_document_size)
            else:
                total_queue_size = buffer_size

            self.single_queue_size = int(total_queue_size / 2)

        tqdm.write(f"Max number of documents in queue: {self.single_queue_size:,}")

        # mp queues are thread-safe while being able to be used across processes
        self.tq: mp.JoinableQueue = mp.JoinableQueue(maxsize=self.single_queue_size)
        self.pq: mp.JoinableQueue = mp.JoinableQueue(maxsize=self.single_queue_size)
        self.func = func

        self.pull_tqdm_kwargs = dict(
            leave=True,
            disable=(not (show_pull_progress_bar and show_progress_bar)),
        )
        if not self.pull_tqdm_kwargs["disable"]:
            self.pull_tqdm_kwargs["position"] = 0  # type: ignore

        self.transform_tqdm_kwargs = dict(
            leave=True,
            disable=(not (show_transform_progress_bar and show_progress_bar)),
        )
        if not self.transform_tqdm_kwargs["disable"]:
            self.transform_tqdm_kwargs["position"] = (
                self.pull_tqdm_kwargs.get("position", 0) + 1  # type: ignore
            )

        self.push_tqdm_kwargs = dict(
            leave=True,
            disable=(not (show_push_progress_bar and show_progress_bar)),
        )
        if not self.push_tqdm_kwargs["disable"]:
            self.push_tqdm_kwargs["position"] = (
                self.pull_tqdm_kwargs.get("position", 0)  # type: ignore
                + self.transform_tqdm_kwargs.get("position", 0)
                + 1
            )

        self.run_in_background = run_in_background

        self.failed_frontier: Dict[str, int] = {}
        self.retry_count = retry_count
        self.after_id = after_id

        # time limit is 1hr 50mins if not set, this is to leave 10mins at the end
        # of sagemaker job to send workflow status email
        self.timeout = timeout
        self.timeout_event = threading.Event()

    def _get_average_document_size(self, sample_documents: List[Dict[str, Any]]):
        """
        Get average size of a document in memory.
        Returns size in bytes.
        """
        document_sizes = [
            getsizeof(sample_document) for sample_document in sample_documents
        ]
        return sum(document_sizes) / len(sample_documents)

    def _get_optimal_chunksize(
        self, sample_documents: List[Dict[str, Any]], method: str
    ) -> int:
        """
        Calculates the optimal batch size given a list of sampled documents and constraints in config
        """
        document_size = self._get_average_document_size(sample_documents)
        document_size = document_size / 2**20
        target_chunk_mb = int(self.config.get_option("upload.target_chunk_mb"))
        max_chunk_size = int(self.config.get_option("upload.max_chunk_size"))
        chunksize = int(target_chunk_mb / document_size)
        chunksize = min(chunksize, max_chunk_size)
        tqdm.write(f"{method.capitalize()} Chunksize: {chunksize}")
        return chunksize

    def _pull(self):
        """
        Iteratively pulls documents from a dataset and places them in the transform queue
        """
        documents: List[Dict[str, Any]] = [{"placeholder": "placeholder"}]
        after_id: Union[List[str], None] = self.after_id

        while not self.timeout_event.is_set():
            current_count = self.pull_bar.n

            if self.pull_chunksize is None:
                # this is the first batch
                pull_chunksize = 20
            else:
                # optimized pull batch size (every other pull)
                pull_chunksize = self.pull_chunksize

            if self.pull_limit is None:
                # Very large number if no limits
                pull_limit = sys.maxsize
            else:
                # if there is a limit, get the ndocs left
                pull_limit = self.pull_limit - current_count

            # Consider all scenarios
            # let 3333 = pull_limit, current_count = 3000
            #         = min(20,            sys.maxsize)             = 20
            #         = min(20,            (3333 - 3000 = 333))     = 20
            #         = min(~512,          sys.maxsize)             = 512
            #         = min(~512,          (3333 - 3000 = 333))     = 333
            page_size = min(pull_chunksize, pull_limit)

            res = self.pull_dataset.datasets.documents.get_where(
                dataset_id=self.pull_dataset_id,
                page_size=page_size,
                filters=self.filters,
                select_fields=self.select_fields,
                sort=[],
                after_id=after_id,
            )

            documents = res["documents"]
            if not documents:
                with self.general_lock:
                    self.ndocs = self.pull_count
                self.tq.task_done()
                break
            after_id = res["after_id"]

            if self.pull_chunksize is None:
                self.pull_chunksize = self._get_optimal_chunksize(
                    documents[:10], "pull"
                )

            for document in documents:
                self.tq.put(document)

            with self.general_lock:
                self.pull_bar.update(len(documents))
                self.pull_count += len(documents)

    def _get_transform_batch(self) -> List[Dict[str, Any]]:
        """
        Collects a batches of of size `transform_chunksize` from the transform queue
        """
        batch: List[Dict[str, Any]] = []

        queue = self.tq
        timeout = 5 if not self.update_all_at_once else None

        if self.transform_count == 0 and self.warmup_chunksize is not None:
            chunksize = self.warmup_chunksize
            tqdm.write("Processing Warmup Batch")
        else:
            chunksize = self.transform_chunksize

        while len(batch) < chunksize:
            try:
                document = queue.get(timeout=timeout)
                batch.append(document)
            except:
                break

        return batch

    def _get_push_batch(self) -> List[Dict[str, Any]]:
        """
        Collects a batches of of size `push_chunksize` from the transform queue
        """
        batch: List[Dict[str, Any]] = []

        queue = self.pq
        timeout = 5

        # Calculate optimal batch size
        if self.push_chunksize is None:
            sample_documents = []
            for _ in range(10):
                try:
                    sample_document = queue.get(timeout=timeout)
                except:
                    break
                sample_documents.append(sample_document)

            self.push_chunksize = self._get_optimal_chunksize(sample_documents, "push")
            batch = sample_documents

        chunksize = self.push_chunksize
        while len(batch) < chunksize:
            try:
                document = queue.get(timeout=timeout)
                batch.append(document)
            except:
                break

        return batch

    def _transform(self):
        """
        Updates a batch of documents given an update function.
        After updating, remove all fields that are present in both old and new batches.
        ^ necessary to avoid reinserting stuff that is already in the cloud.
        Then, repeatedly put each document from the processed batch in the push queue
        """
        while self.transform_count < self.ndocs and not self.timeout_event.is_set():
            with self.transform_batch_lock:
                batch = self._get_transform_batch()

            if self.func is not None:
                old_batch = deepcopy(batch)

                if self.func_lock is not None:
                    with self.func_lock:
                        try:
                            new_batch = self.func(
                                batch, *self.func_args, **self.func_kwargs
                            )
                        except Exception as e:
                            print(e)
                            new_batch = batch
                else:
                    try:
                        new_batch = self.func(batch, **self.func_kwargs)
                    except Exception as e:
                        print(e)
                        new_batch = batch

                batch = PullTransformPush._postprocess(new_batch, old_batch)

            for document in batch:
                self.pq.put(document)

            with self.general_lock:
                self.transform_bar.update(len(batch))
                self.transform_count += len(batch)

    def _handle_failed_documents(
        self,
        res: Dict[str, Any],
        batch: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        """
        Handles any documents that failed to upload.
        Does so by collecting by `_id` from the batch, and reinserting them in the push queue.
        """
        # check if there is any failed documents...
        failed_documents = res["response_json"].get("failed_documents", [])

        if failed_documents:
            with self.general_lock:
                self.failed_documents_count += len(failed_documents)
                desc = f"push - failed_documents = {self.failed_documents_count}"
                self.push_bar.set_description(desc)

            # ...find these failed documents within the batch...
            failed_ids = set(map(lambda x: x["_id"], failed_documents))
            failed_documents = [
                document for document in batch if document["_id"] in failed_ids
            ]

            # ...and re add them to the push queue...
            for failed_document in failed_documents:
                _id = failed_document["_id"]
                if _id not in self.failed_frontier:
                    self.failed_frontier[_id] = 0

                # ...only if they have failed less than the retry count
                if self.failed_frontier[_id] < self.retry_count:
                    self.failed_frontier[_id] += 1
                    self.pq.put(failed_document)

        return failed_documents

    @staticmethod
    def _postprocess(
        new_batch: List[Dict[str, Any]],
        old_batch: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        """
        Removes fields from `new_batch` that are present in the `old_keys` list.
        Necessary to avoid bloating the upload payload with unnecesary information.
        """
        batch = []
        for old_document, new_document in zip(old_batch, new_batch):
            document: Dict[str, Any] = {}
            new_fields = Dataset.list_doc_fields(doc=new_document)
            old_fields = Dataset.list_doc_fields(doc=old_document)
            for field in new_fields:
                old_value = Dataset.get_field(
                    field=field,
                    doc=old_document,
                    missing_treatment="return_none",
                )
                new_value = Dataset.get_field(
                    field=field,
                    doc=new_document,
                    missing_treatment="return_none",
                )
                value_diff = old_value != new_value
                if field not in old_fields or value_diff or field == "_id":
                    Dataset.set_field(field=field, doc=document, value=new_value)
            batch.append(document)

        return batch

    @staticmethod
    def _get_updates(batch) -> bool:
        updates = sum(
            [
                len([key for key in document.keys() if key != "_id"])
                for document in batch
            ]
        )
        return True if updates > 0 else False

    def _push(self) -> None:
        """
        Iteratively gather a batch of processed documents and push these to cloud
        """
        while self.push_count < self.ndocs and not self.timeout_event.is_set():
            with self.push_batch_lock:
                batch = self._get_push_batch()

            batch = self.pull_dataset.json_encoder(batch)
            update = PullTransformPush._get_updates(batch)

            if update:
                res = self.push_dataset.datasets.documents.bulk_update(
                    self.push_dataset_id,
                    batch,
                    return_documents=True,
                    ingest_in_background=self.ingest_in_background,
                )
            else:
                res = {
                    "response_json": {},
                    "documents": batch,
                    "status_code": 200,
                }

            failed_documents = self._handle_failed_documents(res, batch)

            with self.general_lock:
                self.push_bar.update(len(batch) - len(failed_documents))
                self.push_count += len(batch) - len(failed_documents)

    def _init_progress_bars(self) -> None:
        """
        Initialise the progress bars for dispay progress on pulling updating and pushing.
        """
        self.pull_count = 0
        self.pull_bar = tqdm(
            desc="pull",
            total=self.ndocs,
            **self.pull_tqdm_kwargs,
        )
        self.transform_count = 0
        self.transform_bar = tqdm(
            range(self.ndocs),
            desc="transform",
            **self.transform_tqdm_kwargs,
        )
        self.push_count = 0
        self.push_bar = tqdm(
            range(self.ndocs),
            desc="push",
            **self.push_tqdm_kwargs,
        )

    def _init_worker_threads(self) -> None:
        """
        Initialise the worker threads for each process
        """
        daemon = True if self.timeout is not None else False
        self.pull_thread = threading.Thread(target=self._pull, daemon=daemon)
        self.transform_threads = [
            threading.Thread(target=self._transform, daemon=daemon)
            for _ in range(self.transform_workers)
        ]
        self.push_threads = [
            threading.Thread(target=self._push, daemon=daemon)
            for _ in range(self.push_workers)
        ]

    def _start_worker_threads(self):
        """
        Start the worker threads
        """

        # Start fetching data from the server
        self.pull_thread.start()
        # Once there is data in the queue, then we start the
        # transform threads
        while self.tq.empty():
            time.sleep(1)
        for thread in self.transform_threads:
            thread.start()
        while self.pq.empty():
            time.sleep(1)
        for thread in self.push_threads:
            thread.start()

    def _join_worker_threads(self, timeout: Optional[int] = None):
        """
        ...and then join them in reversed order.
        """

        # Try to join if not running in background
        if not self.run_in_background:
            self.pull_thread.join(timeout=timeout)
            for thread in self.transform_threads:
                thread.join(timeout=timeout)
            for thread in self.push_threads:
                thread.join(timeout=timeout)

    def _threads_are_alive(self) -> True:
        """
        Poll all active trheads to check if they are alive
        """
        if self.pull_thread.is_alive():
            return True
        for thread in self.transform_threads:
            if thread.is_alive():
                return True
        for thread in self.push_threads:
            if thread.is_alive():
                return True
        return False

    def _run_timer(self):
        start_time = time.time()
        while True:
            current_time = time.time()

            # check if time limit was exceeded
            if (current_time - start_time) >= self.timeout:
                tqdm.write("Time Limit Exceeded")
                with self.general_lock:
                    self.timeout_event.set()
                tqdm.write("Exiting Operation...")
                break

            # or if all the threads have finished
            if not self._threads_are_alive():
                break

            # poll these checks every 1 sec
            time.sleep(1.0)

    def _flush_queues(self, timeout: float = 1e-2):
        """
        Gets all items in both queues to avoid
        BrokenPipeError when calling queue.close()
        """
        while True:
            try:
                self.tq.get(timeout=timeout)
            except:
                break

        while True:
            try:
                self.pq.get(timeout=timeout)
            except:
                break

        self.tq.close()
        self.pq.close()

    def run(self) -> Dict[str, Any]:
        """
        (Main Method)
        Do the pulling, the updating, and of course, the pushing.

        return the _ids of any failed documents
        """

        if self.ndocs > 0:
            self._init_progress_bars()
            self._init_worker_threads()
            self._start_worker_threads()

            if self.timeout is None:
                self._join_worker_threads()

            else:
                self._run_timer()  # Starts the timer
                self._flush_queues()
                # no need to join threads as they are daemons
                # if there is a timeout

        return {
            "timed_out": self.timeout_event.is_set(),
            "failed_documents": list(self.failed_frontier.keys()),
        }


def arguments(cls: Type[PullTransformPush]):
    import inspect

    sig = inspect.signature(cls.__init__)
    return list(sig.parameters)


class OperationRun(TransformBase):
    """
    All functions related to running transforms as an operation on datasets
    """

    def is_chunk_valid(self, chunk):
        return chunk is not None and len(chunk) > 0

    # @abstractmethod
    def post_run(self, dataset, documents, updated_documents):
        pass

    def run(
        self,
        dataset: Dataset,
        batched: Optional[bool] = False,
        chunksize: Optional[int] = None,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        output_fields: Optional[list] = None,
        refresh: bool = False,
        **kwargs,
    ):
        """It takes a dataset, and then it gets all the documents from that dataset. Then it transforms the
        documents and then it upserts the documents.

        Parameters
        ----------
        dataset : Dataset
            Dataset,
        select_fields : list
            Used to determine which fields to retrieve for filters
        output_fields: list
            Used to determine which output fields are missing to continue running operation

        filters : list
            list = None,

        """

        if filters is None:
            filters = []
        if select_fields is None:
            select_fields = []

        # store this
        if hasattr(dataset, "dataset_id"):
            self.dataset_id = dataset.dataset_id

        schema = dataset.schema
        self._check_fields_in_schema(schema=schema, fields=select_fields)

        filters += [
            {
                "filter_type": "or",
                "condition_value": [
                    {
                        "field": field,
                        "filter_type": "exists",
                        "condition": "==",
                        "condition_value": " ",
                    }
                    for field in select_fields
                ],
            }
        ]

        # add a checkmark for output fields
        if not refresh and output_fields is not None and len(output_fields) > 0:
            filters += [
                {
                    "field": output_fields[0],
                    "filter_type": "exists",
                    "condition": "!=",
                    "condition_value": " ",
                }
            ]

        # needs to be here due to circular imports
        from relevanceai.operations_new.ops_manager import OperationManager

        with OperationManager(
            dataset=dataset,
            operation=self,
        ) as dataset:
            res = self.batch_transform_upsert(
                dataset=dataset,
                select_fields=select_fields,
                filters=filters,
                chunksize=chunksize,
                update_all_at_once=(not batched),
                **kwargs,
            )

        return res

    def batch_transform_upsert(
        self,
        dataset: Dataset,
        func_args: Optional[Tuple[Any]] = None,
        func_kwargs: Optional[Dict[str, Any]] = None,
        select_fields: list = None,
        filters: list = None,
        chunksize: int = None,
        transform_workers: int = 2,
        push_workers: int = 2,
        timeout: int = 30,
        buffer_size: int = 0,
        show_progress_bar: bool = True,
        warmup_chunksize: int = None,
        transform_chunksize: int = 32,
        multithreaded_update: bool = False,
        update_all_at_once: bool = False,
        ingest_in_background: bool = True,
        **kwargs,
    ):
        if multithreaded_update:
            warnings.warn(
                "Multithreaded-update should be False for vectorizing with 1 GPU only. Could hang if True. Works fine on CPU."
            )
        ptp = PullTransformPush(
            dataset=dataset,
            func=self.transform,
            func_args=func_args,
            func_kwargs=func_kwargs,
            multithreaded_update=multithreaded_update,
            pull_chunksize=chunksize,
            warmup_chunksize=warmup_chunksize,
            transform_chunksize=transform_chunksize,
            push_chunksize=chunksize,
            filters=filters,
            select_fields=select_fields,
            transform_workers=transform_workers,
            push_workers=push_workers,
            buffer_size=buffer_size,
            show_progress_bar=show_progress_bar,
            timeout=timeout,
            update_all_at_once=update_all_at_once,
            ingest_in_background=ingest_in_background,
            **kwargs,
        )
        ptp.run()

    def store_operation_metadata(
        self,
        dataset: Dataset,
        values: Optional[Dict[str, Any]] = None,
    ):
        """This function stores metadata about operators

        Parameters
        ----------
        dataset : Dataset
            Dataset,
        values : Optional[Dict[str, Any]]
            Optional[Dict[str, Any]] = None,

        Returns
        -------
            The dataset object with the metadata appended to it.

        .. code-block::

            {
                "_operationhistory_": {
                    "1-1-1-17-2-3": {
                        "operation": "vector",
                        "model_name": "miniLm"
                    },
                }
            }

        """
        if values is None:
            values = self.get_operation_metadata()

        tqdm.write("Storing operation metadata...")
        timestamp = str(datetime.now().timestamp()).replace(".", "-")
        metadata = dataset.metadata.to_dict()
        if "_operationhistory_" not in metadata:
            metadata["_operationhistory_"] = {}
        metadata["_operationhistory_"].update(
            {
                timestamp: {
                    "operation": self.name,
                    "parameters": str(values),
                }
            }
        )
        # Gets metadata and appends to the operation history
        return dataset.upsert_metadata(metadata)



"""
Batch Cluster Operations
"""
from typing import Optional

from relevanceai.dataset import Dataset
from relevanceai.operations_new.cluster.ops import ClusterOps
from relevanceai.operations_new.cluster.batch.transform import BatchClusterTransform
from relevanceai.operations_new.cluster.batch.models.base import BatchClusterModelBase

from relevanceai.constants.links import EXPLORER_APP_LINK
from relevanceai.operations_new.ops_manager import OperationManager
from relevanceai.operations_new.ops_run import PullTransformPush


class BatchClusterOps(BatchClusterTransform, ClusterOps):
    """Batch Clustering related Operations"""

    def __init__(
        self,
        vector_fields,
        alias: str = None,
        model: BatchClusterModelBase = None,
        model_kwargs=None,
        dataset_id: str = None,
        cluster_field="_cluster_",
        verbose: bool = False,
        **kwargs
    ):
        if len(vector_fields) > 1:
            raise NotImplementedError(
                "Currently we do not support more than 1 vector field."
            )

        if dataset_id is not None:
            self.dataset_id = dataset_id
        self.vector_fields = vector_fields
        self.cluster_field = cluster_field
        self.verbose = verbose
        if isinstance(model, str):
            self.model_name = model
        else:
            self.model_name = str(model)

        if model_kwargs is None:
            model_kwargs = {}

        self.model_kwargs = model_kwargs

        for k, v in kwargs.items():
            setattr(self, k, v)

        super().__init__(
            vector_fields=vector_fields,
            model="MiniBatchKmeans" if model is None else model,
            model_kwargs=model_kwargs,
            **kwargs
        )

        self.alias = self._get_alias(alias)

    def fit(self, chunk):
        vectors = self.get_field_across_documents(
            self.vector_fields[0], chunk, missing_treatment="skip"
        )
        if hasattr(self.model, "n_clusters"):
            if len(vectors) < self.model.n_clusters:
                return chunk
        self.model.partial_fit(vectors)
        return chunk

    def run(
        self,
        dataset: Dataset,
        filters: list = None,
        chunksize: Optional[int] = None,
        **kwargs
    ):
        """
        Run batch clustering
        """
        from tqdm.auto import tqdm

        with OperationManager(dataset=dataset, operation=self) as dataset:
            tqdm.write("\nFitting Model...")
            ptp = PullTransformPush(
                dataset=dataset,
                func=self.fit,
                pull_chunksize=chunksize,
                push_chunksize=chunksize,
                filters=filters,
                select_fields=self.vector_fields,
                show_progress_bar=True,
                **kwargs
            )
            ptp.run()

            kwargs.pop("transform_chunksize")

            tqdm.write("\nPredicting Documents...")
            ptp = PullTransformPush(
                dataset=dataset,
                func=self.transform,
                pull_chunksize=chunksize,
                push_chunksize=chunksize,
                filters=filters,
                select_fields=self.vector_fields,
                show_progress_bar=True,
                **kwargs
            )
            ptp.run()

        tqdm.write("\nConfigure your new explore app below:")
        tqdm.write(EXPLORER_APP_LINK.format(dataset.dataset_id))
        return



"""Extract noun operations
"""
import string
from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations_new.ops_base import OperationAPIBase
from relevanceai.operations_new.processing.text.extract_nouns.transform import (
    ExtractNounsTransform,
)


class ExtractNounsOps(ExtractNounsTransform, OperationAPIBase):
    def __init__(
        self,
        fields: list,
        model_name: str,
        output_fields: list,
        cutoff_probability: float,
        stopwords: list = None,
        **kwargs
    ):
        try:
            from flair.models import SequenceTagger
        except ModuleNotFoundError:
            raise MissingPackageError("flair")
        self.fields = fields
        self.model_name = model_name
        self.output_fields = output_fields
        self.cutoff_probability = cutoff_probability
        self.tagger = SequenceTagger.load(model_name)
        from relevanceai.constants import STOPWORDS

        PUNCTUATION = [x for x in string.punctuation]
        if stopwords is not None:
            self.stopwords = STOPWORDS + PUNCTUATION + stopwords
        else:
            self.stopwords = STOPWORDS + PUNCTUATION

        for k, v in kwargs.items():
            setattr(self, k, v)

        # super().__init__(**kwargs)

    def get_operation_metadata(self):
        return {
            "fields": self.fields,
            "model_name": self.model_name,
            "output_fields": self.output_fields,
            "cutoff_probability": self.cutoff_probability,
            "stopwords": self.stopwords,
        }
